import wf_core_data.utils
import requests
import pandas as pd
from collections import OrderedDict
# import pickle
# import json
import datetime
import time
import logging
# import os

logger = logging.getLogger(__name__)

DEFAULT_DELAY = 0.25
DEFAULT_MAX_REQUESTS = 50

SCHOOLS_BASE_ID = 'appJBT9a4f3b7hWQ2'

class AirtableClient:
    def __init__(
        self,
        api_key=None,
        url_base='https://api.airtable.com/v0/'
    ):
        self.api_key = api_key
        self.url_base = url_base
        if self.api_key is None:
            self.api_key = os.getenv('AIRTABLE_API_KEY')

    def fetch_tl_data(
        self,
        pull_datetime=None,
        params=None,
        base_id=SCHOOLS_BASE_ID,
        format='dataframe',
        delay=DEFAULT_DELAY,
        max_requests=DEFAULT_MAX_REQUESTS
    ):
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        logger.info('Fetching TL data from Airtable')
        records = self.bulk_get(
            base_id=base_id,
            endpoint='TLs',
            params=params
        )
        tl_data=list()
        for record in records:
            fields = record.get('fields', {})
            datum = OrderedDict([
                ('teacher_id_at', record.get('id')),
                ('teacher_created_datetime_at', wf_core_data.utils.to_datetime(record.get('createdTime'))),
                ('pull_datetime', pull_datetime),
                ('teacher_full_name_at', fields.get('Full Name')),
                ('teacher_first_name_at', fields.get('First Name')),
                ('teacher_middle_name_at', fields.get('Middle Name')),
                ('teacher_last_name_at', fields.get('Last Name')),
                ('teacher_title_at', fields.get('Title')),
                ('teacher_ethnicity_at', fields.get('Race & Ethnicity')),
                ('teacher_ethnicity_other_at', fields.get('Race & Ethnicity - Other')),
                ('teacher_income_background_at', fields.get('Income Background')),
                ('teacher_email_at', fields.get('Email')),
                ('teacher_email_2_at', fields.get('Email 2')),
                ('teacher_email_3_at', fields.get('Email 3')),
                ('teacher_phone_at', fields.get('Phone Number')),
                ('teacher_phone_2_at', fields.get('Phone Number 2')),
                ('teacher_employer_at', fields.get('Employer')),
                ('hub_at', fields.get('Hub')),
                ('pod_at', fields.get('Pod')),
                ('user_id_tc', fields.get('TC User ID'))
            ])
            tl_data.append(datum)
        if format == 'dataframe':
            tl_data = convert_tl_data_to_df(tl_data)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return tl_data

    def fetch_location_data(
        self,
        pull_datetime=None,
        params=None,
        base_id=SCHOOLS_BASE_ID,
        format='dataframe',
        delay=DEFAULT_DELAY,
        max_requests=DEFAULT_MAX_REQUESTS
    ):
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        logger.info('Fetching location data from Airtable')
        records = self.bulk_get(
            base_id=base_id,
            endpoint='Locations',
            params=params
        )
        location_data=list()
        for record in records:
            fields = record.get('fields', {})
            datum = OrderedDict([
                ('location_id_at', record.get('id')),
                ('location_created_datetime_at', wf_core_data.utils.to_datetime(record.get('createdTime'))),
                ('pull_datetime', pull_datetime),
                ('location_address_at', fields.get('Address')),
                ('school_id_at', wf_core_data.utils.to_singleton(fields.get('School Name'))),
                ('school_location_start_at', wf_core_data.utils.to_date(fields.get('Start of time at location'))),
                ('school_location_end_at', wf_core_data.utils.to_date(fields.get('End of time at location')))
            ])
            location_data.append(datum)
        if format == 'dataframe':
            location_data = convert_location_data_to_df(location_data)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return location_data

    def fetch_teacher_school_data(
        self,
        pull_datetime=None,
        params=None,
        base_id=SCHOOLS_BASE_ID,
        format='dataframe',
        delay=DEFAULT_DELAY,
        max_requests=DEFAULT_MAX_REQUESTS
    ):
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        logger.info('Fetching teacher school association data from Airtable')
        records = self.bulk_get(
            base_id=base_id,
            endpoint='Teachers x Schools',
            params=params
        )
        teacher_school_data=list()
        for record in records:
            fields = record.get('fields', {})
            datum = OrderedDict([
                ('teacher_school_id_at', record.get('id')),
                ('teacher_school_created_datetime_at', wf_core_data.utils.to_datetime(record.get('createdTime'))),
                ('pull_datetime', pull_datetime),
                ('teacher_id_at', fields.get('TL')),
                ('school_id_at', fields.get('School')),
                ('teacher_school_start_at', wf_core_data.utils.to_date(fields.get('Start Date'))),
                ('teacher_school_end_at', wf_core_data.utils.to_date(fields.get('End Date'))),
                ('teacher_school_active_at', wf_core_data.utils.to_boolean(fields.get('Currently Active')))
            ])
            teacher_school_data.append(datum)
        if format == 'dataframe':
            teacher_school_data = convert_teacher_school_data_to_df(teacher_school_data)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return teacher_school_data

    def fetch_school_data(
        self,
        pull_datetime=None,
        params=None,
        base_id=SCHOOLS_BASE_ID,
        format='dataframe',
        delay=DEFAULT_DELAY,
        max_requests=DEFAULT_MAX_REQUESTS
    ):
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        logger.info('Fetching school data from Airtable')
        records = self.bulk_get(
            base_id=base_id,
            endpoint='Schools',
            params=params
        )
        school_data=list()
        for record in records:
            fields = record.get('fields', {})
            datum = OrderedDict([
                ('school_id_at', record.get('id')),
                ('school_created_datetime_at', wf_core_data.utils.to_datetime(record.get('createdTime'))),
                ('pull_datetime', pull_datetime),
                ('hub_at', fields.get('Hub')),
                ('pod_at', fields.get('Pod')),
                ('school_name_at', fields.get('Name')),
                ('school_status_at', fields.get('Status')),
                ('school_ssj_stage_at', fields.get('SSJ Stage')),
                ('school_governance_model_at', fields.get('Governance Model')),
                ('school_ages_served_at', fields.get('Ages served')),
                ('school_phone_number_at', fields.get('Phone Number')),
                ('school_location_ids_at', fields.get('Locations')),
                ('school_time_zone_at', fields.get('Time Zone')),
                ('school_id_tc', fields.get('TC_school_ID'))
            ])
            school_data.append(datum)
        if format == 'dataframe':
            school_data = convert_school_data_to_df(school_data)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return school_data

    def bulk_get(
        self,
        base_id,
        endpoint,
        params=None,
        delay=DEFAULT_DELAY,
        max_requests=DEFAULT_MAX_REQUESTS
    ):
        if params is None:
            params = dict()
        num_requests = 0
        records = list()
        while True:
            data = self.get(
                base_id=base_id,
                endpoint=endpoint,
                params=params
            )
            if 'records' in data.keys():
                logging.info('Returned {} records'.format(len(data.get('records'))))
                records.extend(data.get('records'))
            num_requests += 1
            if num_requests >= max_requests:
                logger.warning('Reached maximum number of requests ({}). Terminating.'.format(
                    max_requests
                ))
                break
            offset = data.get('offset')
            if offset is None:
                break
            params['offset'] = offset
            time.sleep(delay)
        return records

    def post(
        self,
        base_id,
        endpoint,
        data
    ):
        headers = dict()
        if self.api_key is not None:
            headers['Authorization'] = 'Bearer {}'.format(self.api_key)
        r = requests.post(
            '{}{}/{}'.format(
                self.url_base,
                base_id,
                endpoint
            ),
            headers=headers,
            json=data
        )
        if r.status_code != 200:
            error_message = 'Airtable POST request returned status code {}'.format(r.status_code)
            r.raise_for_status()
        return r.json()

    def get(
        self,
        base_id,
        endpoint,
        params=None
    ):
        headers = dict()
        if self.api_key is not None:
            headers['Authorization'] = 'Bearer {}'.format(self.api_key)
        r = requests.get(
            '{}{}/{}'.format(
                self.url_base,
                base_id,
                endpoint
            ),
            params=params,
            headers=headers
        )
        if r.status_code != 200:
            error_message = 'Airtable GET request returned status code {}'.format(r.status_code)
            r.raise_for_status()
        return r.json()

def convert_tl_data_to_df(tl_data):
    if len(tl_data) == 0:
        return pd.DataFrame()
    tl_data_df = pd.DataFrame(
        tl_data,
        dtype='object'
    )
    tl_data_df['pull_datetime'] = pd.to_datetime(tl_data_df['pull_datetime'])
    tl_data_df['teacher_created_datetime_at'] = pd.to_datetime(tl_data_df['teacher_created_datetime_at'])
    # school_data_df['user_id_tc'] = pd.to_numeric(tl_data_df['user_id_tc']).astype('Int64')
    tl_data_df = tl_data_df.astype({
        'teacher_full_name_at': 'string',
        'teacher_middle_name_at': 'string',
        'teacher_last_name_at': 'string',
        'teacher_title_at': 'string',
        'teacher_ethnicity_at': 'string',
        'teacher_ethnicity_other_at': 'string',
        'teacher_income_background_at': 'string',
        'teacher_email_at': 'string',
        'teacher_email_2_at': 'string',
        'teacher_email_3_at': 'string',
        'teacher_phone_at': 'string',
        'teacher_phone_2_at': 'string',
        'teacher_employer_at': 'string',
        'hub_at': 'string',
        'pod_at': 'string',
        'user_id_tc': 'string'
    })
    tl_data_df.set_index('teacher_id_at', inplace=True)
    return tl_data_df

def convert_location_data_to_df(location_data):
    if len(location_data) == 0:
        return pd.DataFrame()
    location_data_df = pd.DataFrame(
        location_data,
        dtype='object'
    )
    location_data_df['pull_datetime'] = pd.to_datetime(location_data_df['pull_datetime'])
    location_data_df['location_created_datetime_at'] = pd.to_datetime(location_data_df['location_created_datetime_at'])
    location_data_df = location_data_df.astype({
        'location_id_at': 'string',
        'location_address_at': 'string',
        'school_id_at': 'string'
    })
    location_data_df.set_index('location_id_at', inplace=True)
    return location_data_df

def convert_teacher_school_data_to_df(teacher_school_data):
    if len(teacher_school_data) == 0:
        return pd.DataFrame()
    teacher_school_data_df = pd.DataFrame(
        teacher_school_data,
        dtype='object'
    )
    teacher_school_data_df['pull_datetime'] = pd.to_datetime(teacher_school_data_df['pull_datetime'])
    teacher_school_data_df['teacher_school_created_datetime_at'] = pd.to_datetime(teacher_school_data_df['teacher_school_created_datetime_at'])
    teacher_school_data_df = teacher_school_data_df.astype({
        'teacher_school_active_at': 'bool'
    })
    teacher_school_data_df.set_index('teacher_school_id_at', inplace=True)
    return teacher_school_data_df

def convert_school_data_to_df(school_data):
    if len(school_data) == 0:
        return pd.DataFrame()
    school_data_df = pd.DataFrame(
        school_data,
        dtype='object'
    )
    school_data_df['pull_datetime'] = pd.to_datetime(school_data_df['pull_datetime'])
    school_data_df['school_created_datetime_at'] = pd.to_datetime(school_data_df['school_created_datetime_at'])
    school_data_df['school_id_tc'] = pd.to_numeric(school_data_df['school_id_tc']).astype('Int64')
    school_data_df = school_data_df.astype({
        'school_id_at': 'string',
        'hub_at': 'string',
        'pod_at': 'string',
        'school_name_at': 'string',
        'school_status_at': 'string',
        'school_ssj_stage_at': 'string',
        'school_governance_model_at': 'string',
        'school_phone_number_at': 'string',
        'school_time_zone_at': 'string',
    })
    school_data_df.set_index('school_id_at', inplace=True)
    return school_data_df



import wf_core_data.rosters.shared_constants
import wf_core_data.rosters.shared_functions
import pandas as pd
import os
import logging

logger = logging.getLogger(__name__)

FASTBRIDGE_TARGET_COLUMN_NAMES = [
    'State',
    'SchoolDistrict',
    'School',
    'Grade',
    'Course',
    'Section',
    'StudentID',
    'StudentStateID',
    'StudentFirstName',
    'StudentLastName',
    'TeacherID',
    'TeacherFirstName',
    'TeacherLastName',
    'TeacherEmail',
    'StudentGender',
    'StudentBirthDate',
    'StudentRace',
    'MealStatus',
    'EnglishProficiency',
    'NativeLanguage',
    'ServiceCode',
    'PrimaryDisabilityType',
    'IEPReading',
    'IEPMath',
    'IEPBehavior',
    'GiftedAndTalented',
    'Section504',
    'Mobility'
]

FASTBRIDGE_GENDER_MAP = {
    'M': 'M',
    'F': 'F',
    'unmatched_value': None,
    'na_value': None
}

FASTBRIDGE_ETHNICITY_MAP = {
    'african_american': 'AA',
    'asian_american': 'AS',
    'hispanic': 'HI',
    'middle_eastern': 'OT',
    'native_american': 'AI',
    'other': 'OT',
    'pacific_islander': 'NH',
    'white': 'WH',
    'unmatched_value': 'OT',
    'na_value': None,
    'multiple_values': 'MT'
}

FASTBRIDGE_GRADE_NAME_MAP = {
    'EC': 'EC',
    'PK': 'PK',
    'PK_3': 'PK',
    'PK_4': 'PK',
    'K': 'KG',
    '1': '01',
    '2': '02',
    '3': '03',
    '4': '04',
    '5': '05',
    '6': '06',
    '7': '07',
    '8': '08',
    '9': '09',
    '10': '10',
    '11': '11',
    '12': '12',
    'unmatched_value': None,
    'na_value': None
}

FASTBRIDGE_TESTABLE_GRADES = [
    'PK',
    'KG',
    '01',
    '02',
    '03',
    '04',
    '05',
    '06',
    '07',
    '08',
    '09',
    '10',
    '11',
    '12'
]

def create_fastbridge_roster_and_write_locally(
    base_directory,
    filename_suffix,
    master_roster_subdirectory='master_rosters',
    master_roster_filename_stem='master_roster',
    fastbridge_roster_subdirectory='fastbridge_rosters',
    fastbridge_roster_filename_stem='fastbridge_roster'
):
    filename = os.path.join(
        base_directory,
        master_roster_subdirectory,
        '{}_{}'.format(
            master_roster_filename_stem,
            filename_suffix
        ),
        '{}_{}.pkl'.format(
            master_roster_filename_stem,
            filename_suffix
        )
    )
    master_roster_data = pd.read_pickle(filename)
    fastbridge_roster_data = wf_core_data.create_fastbridge_roster(
        master_roster_data=master_roster_data
    )
    wf_core_data.write_fastbridge_rosters_local(
        fastbridge_roster_data=fastbridge_roster_data,
        base_directory=base_directory,
        subdirectory=fastbridge_roster_subdirectory,
        filename_stem=fastbridge_roster_filename_stem,
        filename_suffix=filename_suffix
    )

def create_fastbridge_roster(
    master_roster_data
):
    ## Rename fields
    logger.info('Renaming fields')
    fastbridge_roster_data = (
        master_roster_data
        .rename(columns = {
            'school_state': 'State',
            'legal_entity_name_wf': 'SchoolDistrict',
            'school_name_tc': 'School',
            'classroom_name_tc': 'Course',
            'student_id_alt_normalized_tc': 'StudentStateID',
            'student_first_name_tc': 'StudentFirstName',
            'student_last_name_tc': 'StudentLastName',
            'teacher_id_tc': 'TeacherID',
            'teacher_first_name_tc': 'TeacherFirstName',
            'teacher_last_name_tc': 'TeacherLastName',
            'teacher_email_tc':  'TeacherEmail'
        })
    )
    ## Create new fields
    ### Section
    fastbridge_roster_data['Section'] = 'S1'
    ### Student ID
    logger.info('Creating student ID field')
    fastbridge_roster_data['StudentID'] = fastbridge_roster_data.index.get_level_values('student_id_tc')
    ### Student birth date
    logger.info('Creating birth date field')
    fastbridge_roster_data['StudentBirthDate'] = fastbridge_roster_data['student_birth_date_tc'].apply(
        lambda x: x.strftime('%m/%d/%Y')
    )
    ### Student gender
    logger.info('Creating gender field')
    fastbridge_roster_data['StudentGender'] = fastbridge_roster_data['student_gender_wf'].apply(
        lambda x: FASTBRIDGE_GENDER_MAP.get(x, FASTBRIDGE_GENDER_MAP.get('unmatched_value')) if pd.notna(x) else FASTBRIDGE_GENDER_MAP.get('na_value')
    )
    ### Grade
    logger.info('Creating grade field')
    fastbridge_roster_data['Grade'] = fastbridge_roster_data['student_grade_wf'].apply(
        lambda x: FASTBRIDGE_GRADE_NAME_MAP.get(x, FASTBRIDGE_GRADE_NAME_MAP.get('unmatched_value')) if pd.notna(x) else FASTBRIDGE_GRADE_NAME_MAP.get('na_value')
    )
    ### Student ethnicity
    logger.info('Creating ethnicity field')
    def student_race_fastbridge(ethnicity_list):
        if not isinstance(ethnicity_list, list):
            return FASTBRIDGE_ETHNICITY_MAP.get('na_value')
        if len(ethnicity_list) > 1:
            return FASTBRIDGE_ETHNICITY_MAP.get('multiple_values')
        return FASTBRIDGE_ETHNICITY_MAP.get(ethnicity_list[0], FASTBRIDGE_ETHNICITY_MAP.get('unmatched_value'))
    fastbridge_roster_data['StudentRace'] = fastbridge_roster_data['student_ethnicity_wf'].apply(student_race_fastbridge)
    ## Arrange columns and rows
    logger.info('Rearranging columns and rows')
    fastbridge_roster_data = (
        fastbridge_roster_data
        .reindex(columns=(
            wf_core_data.rosters.shared_constants.GROUPING_COLUMN_NAMES +
            FASTBRIDGE_TARGET_COLUMN_NAMES
        ))
        .sort_values(
            wf_core_data.rosters.shared_constants.GROUPING_COLUMN_NAMES +
            ['Grade', 'StudentFirstName', 'StudentLastName']
        )
    )
    ## Create output
    logger.info('Restriction to testable grades. {} student records before restricting'.format(
        len(fastbridge_roster_data)
    ))
    fastbridge_roster_data = (
        fastbridge_roster_data
        .loc[fastbridge_roster_data['Grade'].isin(FASTBRIDGE_TESTABLE_GRADES)]
        .copy()
        .reset_index(drop=True)
        .astype('object')
    )
    logger.info('Restricted to testable grades. {} student records after restricting'.format(
        len(fastbridge_roster_data)
    ))
    return fastbridge_roster_data

def write_fastbridge_rosters_local(
    fastbridge_roster_data,
    base_directory,
    subdirectory='fastbridge_rosters',
    filename_stem='fastbridge_roster',
    filename_suffix=None

):
    wf_core_data.rosters.shared_functions.write_rosters_local(
        roster_data=fastbridge_roster_data,
        base_directory=base_directory,
        subdirectory=subdirectory,
        filename_stem=filename_stem,
        filename_suffix=filename_suffix

    )



import pandas as pd

def to_datetime(object):
    try:
        datetime = pd.to_datetime(object, utc=True).to_pydatetime()
        if pd.isnull(datetime):
            date = None
    except:
        datetime = None
    return datetime

def to_date(object):
    try:
        date = pd.to_datetime(object).date()
        if pd.isnull(date):
            date = None
    except:
        date = None
    return date

def to_singleton(object):
    try:
        num_elements = len(object)
        if num_elements > 1:
            raise ValueError('More than one element in object. Conversion to singleton failed')
        if num_elements == 0:
            return None
        return object[0]
    except:
        return object

def to_boolean(object):
    if isinstance(object, bool):
        return object
    if isinstance(object, str):
        if object in ['True', 'TRUE', 'T']:
            return True
        if object in ['False', 'FALSE', 'F']:
            return False
        return None
    if isinstance(object, int):
        if object == 1:
            return True
        if object == 0:
            return False
        return None
    return None

def extract_alphanumeric(object):
    if pd.isna(object):
        return None
    try:
        object_string = str(object)
    except:
        return None
    alphanumeric_string = ''.join(ch for ch in object_string if ch.isalnum())
    return alphanumeric_string

def infer_school_year(
    date,
    rollover_month=7,
    rollover_day=31
):
    if pd.isna(date):
        return None
    if date.month <= rollover_month and date.day <= rollover_day:
        return '{}-{}'.format(
            date.year - 1,
            date.year
        )
    else:
        return '{}-{}'.format(
            date.year,
            date.year + 1
        )

def filter_dataframe(
    dataframe,
    filter_dict=None
):
    if filter_dict is None:
        return dataframe
    index_columns = dataframe.index.names
    dataframe=dataframe.reset_index()
    for key, value_list in filter_dict.items():
        dataframe = dataframe.loc[dataframe[key].isin(value_list)]
    dataframe.set_index(index_columns, inplace=True)
    return dataframe

def select_from_dataframe(
    dataframe,
    select_dict=None
):
    if select_dict is None:
        return dataframe
    keys, values = zip(*select_dict.items())
    dataframe = dataframe.xs(values, level=keys)
    return dataframe



import wf_core_data.rosters.shared_constants
import pandas as pd
import os
import logging

logger = logging.getLogger(__name__)

def write_rosters_local(
    roster_data,
    base_directory,
    subdirectory,
    filename_stem,
    filename_suffix=None

):
    if filename_suffix is None:
        filename_suffix = datetime.datetime.now(tz=datetime.timezone.utc).strftime('%Y%m%d')
    logger.info('Filename suffix is {}'.format(filename_suffix))
    output_directory_base = os.path.join(
        base_directory,
        subdirectory,
        '{}_{}'.format(
            filename_stem,
            filename_suffix
        )
    )
    output_directory_csv = os.path.join(
        output_directory_base,
        'csv'
    )
    output_directory_pickle = os.path.join(
        output_directory_base,
        'pickle'
    )
    output_directory_excel = os.path.join(
        output_directory_base,
        'excel'
    )
    os.makedirs(output_directory_csv, exist_ok=True)
    os.makedirs(output_directory_pickle, exist_ok=True)
    os.makedirs(output_directory_excel, exist_ok=True)
    output = (
        roster_data
        .drop(columns=wf_core_data.rosters.shared_constants.GROUPING_COLUMN_NAMES)
    )
    filename = '{}_{}'.format(
        filename_stem,
        filename_suffix
    )
    output.to_csv(
        os.path.join(
            output_directory_csv,
            '{}.csv'.format(
                filename
            )
        ),
        index = False
    )
    output.to_pickle(
        os.path.join(
            output_directory_pickle,
            '{}.pkl'.format(
                filename
            )
        )
    )
    output.to_excel(
        os.path.join(
            output_directory_excel,
            '{}.xlsx'.format(
                filename
            )
        )
    )
    for legal_entity_short_name, roster_df_group in roster_data.groupby('legal_entity_short_name_wf'):
        output = (
            roster_df_group
            .drop(columns=wf_core_data.rosters.shared_constants.GROUPING_COLUMN_NAMES)
        )
        filename = '{}_{}_{}'.format(
            filename_stem,
            legal_entity_short_name,
            filename_suffix
        )
        output.to_csv(
            os.path.join(
                output_directory_csv,
                '{}.csv'.format(
                    filename
                )
            ),
            index = False
        )
        output.to_pickle(
            os.path.join(
                output_directory_pickle,
                '{}.pkl'.format(
                    filename
                )
            )
        )
        output.to_excel(
            os.path.join(
                output_directory_excel,
                '{}.xlsx'.format(
                    filename
                )
            )
        )
    for school_short_name, roster_df_group in roster_data.groupby('school_short_name_wf'):
        output = (
            roster_df_group
            .drop(columns=wf_core_data.rosters.shared_constants.GROUPING_COLUMN_NAMES)
        )
        filename = '{}_{}_{}'.format(
            filename_stem,
            school_short_name,
            filename_suffix
        )
        output.to_csv(
            os.path.join(
                output_directory_csv,
                '{}.csv'.format(
                    filename
                )
            ),
            index = False
        )
        output.to_pickle(
            os.path.join(
                output_directory_pickle,
                '{}.pkl'.format(
                    filename
                )
            )
        )
        output.to_excel(
            os.path.join(
                output_directory_excel,
                '{}.xlsx'.format(
                    filename
                )
            ),
            index=False
        )
    for classroom_short_name, roster_df_group in roster_data.groupby('classroom_short_name_wf'):
        output = (
            roster_df_group
            .drop(columns=wf_core_data.rosters.shared_constants.GROUPING_COLUMN_NAMES)
        )
        filename = '{}_{}_{}'.format(
            filename_stem,
            classroom_short_name,
            filename_suffix
        )
        output.to_csv(
            os.path.join(
                output_directory_csv,
                '{}.csv'.format(
                    filename
                )
            ),
            index = False
        )
        output.to_pickle(
            os.path.join(
                output_directory_pickle,
                '{}.pkl'.format(
                    filename
                )
            )
        )
        output.to_excel(
            os.path.join(
                output_directory_excel,
                '{}.xlsx'.format(
                    filename
                )
            ),
            index=False
        )



from .google_sheets import *
from .transparent_classroom import *
from .airtable import *
from .rosters.master_roster import *
from .rosters.nwea_roster import *
from .rosters.fastbridge_roster import *
from .rosters.fountas_pinnell_roster import *
from .rosters.mefs_roster import *
from .rosters.wida_roster import *
from .analysis.fastbridge_analysis import *
from .analysis.nwea_analysis import *
from .utils import *



import wf_core_data.rosters.shared_constants
import wf_core_data.rosters.shared_functions
import pandas as pd
import os
import logging

logger = logging.getLogger(__name__)

WIDA_TARGET_COLUMN_NAMES = [
    'State Name Abbreviation',
    'District Name',
    'District Number',
    'School Name',
    'School Number',
    'Student Last Name',
    'Student First Name',
    'Student Middle Name',
    'Birth Date',
    'Gender',
    'State Student ID',
    'District Student ID',
    'Grade',
    'Ethnicity - Hispanic/Latino',
    'Race - American Indian/Alaskan Native',
    'Race - Asian',
    'Race - Black/African American',
    'Race - Pacific Islander/Hawaiian',
    'Race - White',
    'Native Language',
    'Date First Enrolled US School',
    'Length of Time in LEP/ELL Program',
    'Title III Status',
    'Migrant',
    'IEP Status',
    '504 Plan',
    'Primary Disability',
    'Secondary Disability',
    'LIEP Classification',
    'LIEP - Parental Refusal',
    'LIEP - Optional Data',
    'Accommodation - MC',
    'Accommodation - RA',
    'Accommodation - ES',
    'Accommodation - LP',
    'Accommodation - BR',
    'Accommodation - SD',
    'Accommodation - IR',
    'Accommodation - RP',
    'Accommodation - SR',
    'Accommodation - WD',
    'Accommodation - RD',
    'Accommodation - NS',
    'Accommodation - EM',
    'State Defined Optional Data',
    'District Defined Optional Data',
    'Mode of Administration',
    'Paper Tier',
    'Alternate ACCESS for ELLs Tester',
    'Student Type',
    'Additional field to be used by a state if needed'
]

WIDA_GENDER_MAP = {
    'M': 'M',
    'F': 'F',
    'unmatched_value': None,
    'na_value': None
}

WIDA_GRADE_NAME_MAP = {
    'EC': 'EC',
    'PK': 'PK',
    'PK_3': 'PK',
    'PK_4': 'PK',
    'K': '00',
    '1': '01',
    '2': '02',
    '3': '03',
    '4': '04',
    '5': '05',
    '6': '06',
    '7': '07',
    '8': '08',
    '9': '09',
    '10': '10',
    '11': '11',
    '12': '12',
    'unmatched_value': None,
    'na_value': None
}

WIDA_TESTABLE_GRADES = [
    '00'
]

def create_wida_roster_and_write_locally(
    base_directory,
    filename_suffix,
    master_roster_subdirectory='master_rosters',
    master_roster_filename_stem='master_roster',
    wida_roster_subdirectory='wida_rosters',
    wida_roster_filename_stem='wida_roster'
):
    filename = os.path.join(
        base_directory,
        master_roster_subdirectory,
        '{}_{}'.format(
            master_roster_filename_stem,
            filename_suffix
        ),
        '{}_{}.pkl'.format(
            master_roster_filename_stem,
            filename_suffix
        )
    )
    master_roster_data = pd.read_pickle(filename)
    wida_roster_data = wf_core_data.create_wida_roster(
        master_roster_data=master_roster_data
    )
    wf_core_data.write_wida_rosters_local(
        wida_roster_data=wida_roster_data,
        base_directory=base_directory,
        subdirectory=wida_roster_subdirectory,
        filename_stem=wida_roster_filename_stem,
        filename_suffix=filename_suffix
    )

def create_wida_roster(
    master_roster_data
):
    ## Rename fields
    logger.info('Renaming fields')
    wida_roster_data = (
        master_roster_data
        .rename(columns = {
            'school_state': 'State Name Abbreviation',
            'district_name_wida': 'District Name',
            'district_id_wida': 'District Number',
            'school_name_wida': 'School Name',
            'school_id_wida': 'School Number',
            'student_last_name_tc': 'Student Last Name',
            'student_first_name_tc': 'Student First Name',
            'student_middle_name_tc': 'Student Middle Name',
            'student_id_alt_normalized_tc': 'State Student ID'
        })
    )
    ## Create new fields
    ### Student ID
    logger.info('Creating district student ID field')
    wida_roster_data['District Student ID'] = wida_roster_data.apply(
        lambda row: '{}-{}'.format(
            row.name[0],
            row.name[1]
        ),
        axis=1
    )
    ### Student birth date
    logger.info('Creating birth date field')
    wida_roster_data['Birth Date'] = wida_roster_data['student_birth_date_tc'].apply(
        lambda x: x.strftime('%m/%d/%Y')
    )
    ### Student gender
    logger.info('Creating gender field')
    wida_roster_data['Gender'] = wida_roster_data['student_gender_wf'].apply(
        lambda x: WIDA_GENDER_MAP.get(x, WIDA_GENDER_MAP.get('unmatched_value')) if pd.notna(x) else WIDA_GENDER_MAP.get('na_value')
    )
    ### Grade
    logger.info('Creating grade field')
    wida_roster_data['Grade'] = wida_roster_data['student_grade_wf'].apply(
        lambda x: WIDA_GRADE_NAME_MAP.get(x, WIDA_GRADE_NAME_MAP.get('unmatched_value')) if pd.notna(x) else WIDA_GRADE_NAME_MAP.get('na_value')
    )
    ### Student ethnicity
    logger.info('Creating ethnicity fields')
    wida_roster_data['Ethnicity - Hispanic/Latino'] = wida_roster_data['student_ethnicity_wf'].apply(lambda ethnicity_list:
        'Y' if isinstance(ethnicity_list, list) and 'hispanic' in ethnicity_list else None
    )
    wida_roster_data['Race - American Indian/Alaskan Native'] = wida_roster_data['student_ethnicity_wf'].apply(lambda ethnicity_list:
        'Y' if isinstance(ethnicity_list, list) and 'native_american' in ethnicity_list else None
    )
    wida_roster_data['Race - Asian'] = wida_roster_data['student_ethnicity_wf'].apply(lambda ethnicity_list:
        'Y' if isinstance(ethnicity_list, list) and 'asian_american' in ethnicity_list else None
    )
    wida_roster_data['Race - Black/African American'] = wida_roster_data['student_ethnicity_wf'].apply(lambda ethnicity_list:
        'Y' if isinstance(ethnicity_list, list) and 'african_american' in ethnicity_list else None
    )
    wida_roster_data['Race - Pacific Islander/Hawaiian'] = wida_roster_data['student_ethnicity_wf'].apply(lambda ethnicity_list:
        'Y' if isinstance(ethnicity_list, list) and 'pacific_islander' in ethnicity_list else None
    )
    wida_roster_data['Race - White'] = wida_roster_data['student_ethnicity_wf'].apply(lambda ethnicity_list:
        'Y' if isinstance(ethnicity_list, list) and 'white' in ethnicity_list else None
    )
    ## Arrange columns and rows
    logger.info('Rearranging columns and rows')
    wida_roster_data = (
        wida_roster_data
        .reindex(columns=(
            wf_core_data.rosters.shared_constants.GROUPING_COLUMN_NAMES +
            WIDA_TARGET_COLUMN_NAMES
        ))
        .sort_values(
            wf_core_data.rosters.shared_constants.GROUPING_COLUMN_NAMES +
            ['Grade', 'Student First Name', 'Student Last Name']
        )
    )
    ## Create output
    logger.info('Restriction to testable grades. {} student records before restricting'.format(
        len(wida_roster_data)
    ))
    wida_roster_data = (
        wida_roster_data
        .loc[wida_roster_data['Grade'].isin(WIDA_TESTABLE_GRADES)]
        .copy()
        .reset_index(drop=True)
        .astype('object')
    )
    logger.info('Restricted to testable grades. {} student records after restricting'.format(
        len(wida_roster_data)
    ))
    return wida_roster_data

def write_wida_rosters_local(
    wida_roster_data,
    base_directory,
    subdirectory='wida_rosters',
    filename_stem='wida_roster',
    filename_suffix=None

):
    wf_core_data.rosters.shared_functions.write_rosters_local(
        roster_data=wida_roster_data,
        base_directory=base_directory,
        subdirectory=subdirectory,
        filename_stem=filename_stem,
        filename_suffix=filename_suffix

    )



from wf_core_data.utils import to_date
from gspread_pandas import Spread, Client
import pandas as pd
import json
import datetime
import time
import logging

logger = logging.getLogger(__name__)

def ingest_student_data_google_sheets(sheet_metadata, delay=None):
    df_list = list()
    logger.info('Ingesting data from each sheet')
    for sheet_metadatum in sheet_metadata:
        pull_date = sheet_metadatum['pull_date']
        sheet_id = sheet_metadatum['sheet_id']
        df_sheet = ingest_student_data_google_sheet(sheet_id, pull_date)
        df_list.append(df_sheet)
        if delay is not None:
            logger.info('Waiting {} seconds'.format(delay))
            time.sleep(delay)
    logger.info('Ingested data from {} sheets. Concatenating.'.format(len(df_list)))
    df = pd.concat(df_list, ignore_index=True)
    df.sort_values(['school_id_tc', 'student_id_tc', 'pull_datetime'], inplace=True, ignore_index=True)
    return df

def ingest_student_data_google_sheet(sheet_id, pull_date):
    logger.info('Ingesting data from sheet with pull date {} and ID {}'.format(pull_date, sheet_id))
    spread = Spread(sheet_id)
    df = spread.sheet_to_df(index=None)
    df['school_id_tc'] = pd.to_numeric(df['school_id']).astype('Int64')
    df['child_raw_dict'] = df['child_raw'].apply(lambda x: json.loads(x))
    df['student_id_tc'] = pd.to_numeric(df['child_raw_dict'].apply(lambda x: int(x.get('id')))).astype('Int64')
    df['pull_datetime'] = pd.to_datetime(pull_date, utc=True)
    df['student_first_name_tc'] = df['child_raw_dict'].apply(lambda x: x.get('first_name')).astype('string')
    df['student_middle_name_tc'] = df['child_raw_dict'].apply(lambda x: x.get('middle_name')).astype('string')
    df['student_last_name_tc'] = df['child_raw_dict'].apply(lambda x: x.get('last_name')).astype('string')
    df['student_birth_date_tc'] = df['child_raw_dict'].apply(lambda x: to_date(x.get('birth_date')))
    df['student_gender_tc'] = df['child_raw_dict'].apply(lambda x: x.get('gender')).astype('string')
    df['student_ethnicity_tc'] = df['child_raw_dict'].apply(lambda x: x.get('ethnicity'))
    df['student_dominant_language_tc'] = df['child_raw_dict'].apply(lambda x: x.get('dominant_language')).astype('string')
    df['student_household_income_tc'] = df['child_raw_dict'].apply(lambda x: x.get('household_income')).astype('string')
    df['student_grade_tc'] = df['child_raw_dict'].apply(lambda x: x.get('grade')).astype('string')
    df['student_classroom_ids_tc'] = df['child_raw_dict'].apply(lambda x: x.get('classroom_ids'))
    df['student_program_tc'] = df['child_raw_dict'].apply(lambda x: x.get('program')).astype('string')
    df['student_hours_string_tc'] = df['child_raw_dict'].apply(lambda x: x.get('hours_string')).astype('string')
    df['student_id_alt_tc'] = df['child_raw_dict'].apply(lambda x: x.get('student_id')).astype('string')
    df['student_allergies_tc'] = df['child_raw_dict'].apply(lambda x: x.get('allergies')).astype('string')
    df['student_parent_ids_tc'] = df['child_raw_dict'].apply(lambda x: x.get('parent_ids'))
    df['student_approved_adults_string_tc'] = df['child_raw_dict'].apply(lambda x: x.get('approved_adults_string')).astype('string')
    df['student_emergency_contacts_string_tc'] = df['child_raw_dict'].apply(lambda x: x.get('emergency_contacts_string')).astype('string')
    df['student_notes_tc'] = df['child_raw_dict'].apply(lambda x: x.get('notes')).astype('string')
    df['student_last_day_tc'] = df['child_raw_dict'].apply(lambda x: to_date(x.get('last_day')))
    df['student_exit_reason_tc'] = df['child_raw_dict'].apply(lambda x: x.get('exit_reason')).astype('string')
    df['student_exit_survey_id_tc'] = pd.to_numeric(df['child_raw_dict'].apply(lambda x: x.get('exit_survey_id'))).astype('Int64')
    df['student_exit_notes_tc'] = df['child_raw_dict'].apply(lambda x: x.get('exit_notes')).astype('string')
    df = df.reindex(columns=[
        'school_id_tc',
        'student_id_tc',
        'pull_datetime',
        'student_first_name_tc',
        'student_middle_name_tc',
        'student_last_name_tc',
        'student_birth_date_tc',
        'student_gender_tc',
        'student_ethnicity_tc',
        'student_dominant_language_tc',
        'student_household_income_tc',
        'student_grade_tc',
        'student_classroom_ids_tc',
        'student_program_tc',
        'student_hours_string_tc',
        'student_id_alt_tc',
        'student_allergies_tc',
        'student_parent_ids_tc',
        'student_approved_adults_string_tc',
        'student_emergency_contacts_string_tc',
        'student_notes_tc',
        'student_last_day_tc',
        'student_exit_reason_tc',
        'student_exit_survey_id_tc',
        'student_exit_notes_tc'
    ])
    if df.duplicated(['school_id_tc', 'student_id_tc']).any():
        raise ValueError('Ingested data contains duplicate Transparent Classroom school ID/student id combinations')
    return df



import wf_core_data
import pandas as pd
import slugify
import datetime
import re
import os
import logging

logger = logging.getLogger(__name__)

def fetch_master_roster_data_and_write_local(
    base_directory,
    transparent_classroom_client=None,
    transparent_classroom_username=None,
    transparent_classroom_password=None,
    transparent_classroom_api_token=None,
    transparent_classroom_url_base='https://www.transparentclassroom.com/api/v1/',
    hub_info_path_stem = 'hub_info',
    legal_entity_info_path_stem = 'legal_entity_info',
    school_info_path_stem = 'school_info',
    classroom_info_path_stem = 'classroom_info',
    teacher_info_path_stem = 'teacher_info',
    ethnicity_info_path_stem = 'ethnicity_info',
    gender_map_path_stem = 'gender_map',
    ethnicity_map_path_stem = 'ethnicity_map',
    grade_map_path_stem = 'grade_map',
    subdirectory='master_rosters',
    filename_stem='master_roster',
    filename_suffix=None
):
    master_roster_data = fetch_master_roster_data(
        base_directory=base_directory,
        transparent_classroom_client=transparent_classroom_client,
        transparent_classroom_username=transparent_classroom_username,
        transparent_classroom_password=transparent_classroom_password,
        transparent_classroom_api_token=transparent_classroom_api_token,
        transparent_classroom_url_base=transparent_classroom_url_base,
        hub_info_path_stem=hub_info_path_stem,
        legal_entity_info_path_stem=legal_entity_info_path_stem,
        school_info_path_stem=school_info_path_stem,
        classroom_info_path_stem=classroom_info_path_stem,
        teacher_info_path_stem=teacher_info_path_stem,
        ethnicity_info_path_stem=ethnicity_info_path_stem,
        gender_map_path_stem=gender_map_path_stem,
        ethnicity_map_path_stem=ethnicity_map_path_stem,
        grade_map_path_stem=grade_map_path_stem
    )
    write_master_roster_data_local(
        master_roster_data=master_roster_data,
        base_directory=base_directory,
        subdirectory=subdirectory,
        filename_stem=filename_stem,
        filename_suffix=filename_suffix
    )

def fetch_master_roster_data(
    base_directory,
    transparent_classroom_client=None,
    transparent_classroom_username=None,
    transparent_classroom_password=None,
    transparent_classroom_api_token=None,
    transparent_classroom_url_base='https://www.transparentclassroom.com/api/v1/',
    hub_info_path_stem='hub_info',
    legal_entity_info_path_stem='legal_entity_info',
    school_info_path_stem='school_info',
    classroom_info_path_stem='classroom_info',
    teacher_info_path_stem='teacher_info',
    ethnicity_info_path_stem='ethnicity_info',
    gender_map_path_stem='gender_map',
    ethnicity_map_path_stem='ethnicity_map',
    grade_map_path_stem='grade_map'
):
    if transparent_classroom_client is None:
        transparent_classroom_client = wf_core_data.transparent_classroom.TransparentClassroomClient(
            username=transparent_classroom_username,
            password=transparent_classroom_password,
            api_token=transparent_classroom_api_token,
            url_base=transparent_classroom_url_base
        )
    pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
    ## Fetch target entity info
    ### Hubs
    logger.info('Fetching target hub info')
    hubs = pd.read_csv(
        os.path.join(
            base_directory,
            '.'.join([
                hub_info_path_stem,
                'csv'
            ])
        ),
        index_col='hub_short_name_wf'
    )
    logger.info('Fetched info for {} target hubs'.format(len(hubs)))
    ### Legal entities
    logger.info('Fetching target legal entity info')
    legal_entities = pd.read_csv(
        os.path.join(
            base_directory,
            '.'.join([
                legal_entity_info_path_stem,
                'csv'
            ])
        ),
        index_col='legal_entity_short_name_wf'
    )
    logger.info('Fetched info for {} target legal entities'.format(len(legal_entities)))
    ### Schools
    logger.info('Fetching target school info')
    schools = pd.read_csv(
        os.path.join(
            base_directory,
            '.'.join([
                school_info_path_stem,
                'csv'
            ])
        ),
        index_col='school_id_tc',
        dtype='object'
    )
    logger.info('Fetched info for {} target schools'.format(len(schools)))
    ### Classrooms
    logger.info('Fetching target classroom info')
    classrooms = pd.read_csv(
        os.path.join(
            base_directory,
            '.'.join([
                classroom_info_path_stem,
                'csv'
            ])
        ),
        index_col=['school_id_tc', 'classroom_id_tc']
    )
    logger.info('Fetched info for {} target classrooms'.format(len(classrooms)))
    ### Teachers
    logger.info('Fetching target teacher info')
    teachers = pd.read_csv(
        os.path.join(
            base_directory,
            '.'.join([
                teacher_info_path_stem,
                'csv'
            ])
        ),
        index_col=['school_id_tc', 'teacher_id_tc']
    )
    logger.info('Fetched info for {} target teachers'.format(len(teachers)))
    ### Ethnicities
    logger.info('Fetching target ethnicity info')
    ethnicities = pd.read_csv(
        os.path.join(
            base_directory,
            '.'.join([
                ethnicity_info_path_stem,
                'csv'
            ])
        ),
        index_col='ethnicity_short_name_wf'
    )
    logger.info('Fetched info for {} target ethnicities'.format(len(ethnicities)))
    ## Fetch mappings
    ### Ethnicity map
    logger.info('Fetching ethnicity mapping info')
    ethnicity_map = pd.read_csv(
        os.path.join(
            base_directory,
            '.'.join([
                ethnicity_map_path_stem,
                'csv'
            ])
        ),
        index_col='ethnicity_tc'
    )
    logger.info('Fetched mapping info for {} ethnicity values'.format(len(ethnicity_map)))
    ### Gender map
    logger.info('Fetching gender mapping info')
    gender_map = pd.read_csv(
        os.path.join(
            base_directory,
            '.'.join([
                gender_map_path_stem,
                'csv'
            ])
        ),
        index_col='gender_tc'
    )
    logger.info('Fetched mapping info for {} gender values'.format(len(gender_map)))
    ### Grade map
    logger.info('Fetching grade mapping info')
    grade_map = pd.read_csv(
        os.path.join(
            base_directory,
            '.'.join([
                grade_map_path_stem,
                'csv'
            ])
        ),
        index_col='grade_tc_slugified'
    )
    logger.info('Fetched mapping info for {} grade values'.format(len(grade_map)))
    ## Fetch data from Transparent Classroom
    ### School data
    logger.info('Fetching school data from Transparent Classroom')
    school_data  = (
        transparent_classroom_client.fetch_school_data(
            pull_datetime=pull_datetime,
            format='dataframe'
        )
        .join(
            schools,
            how='inner'
        )
    )
    logger.info('Fetched data for {} target schools'.format(len(school_data)))
    ### Classroom data
    logger.info('Fetching classroom data from Transparent Classroom')
    classroom_data = (
        transparent_classroom_client.fetch_classroom_data(
            school_ids=schools.index,
            pull_datetime=pull_datetime,
            format='dataframe'
        )
        .join(
            classrooms,
            how='inner'
        )
    )
    logger.info('Fetched data for {} target classrooms'.format(len(classroom_data)))
    ### Student classroom data
    logger.info('Fetching student classroom association data from Transparent Classroom')
    student_classroom_data = (
        transparent_classroom_client.fetch_student_classroom_data(
            school_ids=schools.index,
            session_data=None,
            pull_datetime=pull_datetime,
            only_current=True,
            format='dataframe'
        )
        .reset_index(level='session_id_tc', drop=True)
        .reset_index(level='classroom_id_tc')
    )
    logger.info('Fetched {} student classroom associations'.format(len(student_classroom_data)))
    ### Teacher data
    logger.info('Fetching teacher data from Transparent Classroom')
    teacher_data = (
        transparent_classroom_client.fetch_teacher_data(
            school_ids=schools.index,
            pull_datetime=pull_datetime,
            format='dataframe'
        )
        .rename_axis(index={'user_id_tc': 'teacher_id_tc'})
        .join(
            teachers,
            how='inner'
        )
        .rename(columns={
            'user_first_name_tc': 'teacher_first_name_tc',
            'user_last_name_tc': 'teacher_last_name_tc',
            'user_email_tc': 'teacher_email_tc'
        })
    )
    logger.info('Fetched data for {} target teachers'.format(len(teacher_data)))
    ### Student data
    logger.info('Fetching student data from Transparent Classroom')
    student_data, student_parent_data = transparent_classroom_client.fetch_student_data(
        school_ids=schools.index,
        pull_datetime=pull_datetime,
        only_current=True,
        format='dataframe'
    )
    logger.info('Fetched data for {} students'.format(len(student_data)))
    ## Join data tables
    logger.info('Joining student data with other data')
    master_roster_data = (
        student_data
        .join(
            student_classroom_data.drop(columns='pull_datetime'),
            how='left'
        )
        .join(
            classroom_data.drop(columns='pull_datetime'),
            how='inner',
            on=['school_id_tc', 'classroom_id_tc']
        )
        .join(
            school_data.drop(columns='pull_datetime'),
            how='left',
            on='school_id_tc'
        )
        .join(
            legal_entities,
            how='left',
            on='legal_entity_short_name_wf'
        )
        .join(
            hubs,
            how='left',
            on='hub_short_name_wf'
        )
        .join(
            teacher_data.drop(columns='pull_datetime'),
            how='left',
            on=['school_id_tc', 'teacher_id_tc']
        )
    )
    logger.info('Roster includes {} students after joining with other data'.format(len(master_roster_data)))
    ## Construct new fields
    ### School ZIP code
    logger.info('Constructing school ZIP code field')
    zip_re = re.compile(r' (?P<zip>[0-9]{5})')
    master_roster_data['school_zip_code_tc'] = master_roster_data['school_address_tc'].apply(
        lambda x: zip_re.search(x).group('zip') if zip_re.search(x) is not None else None
    )
    ### Normalized alternative student ID (typically state ID)
    logger.info('Constructing normalized alternative student ID field')
    master_roster_data['student_id_alt_normalized_tc'] = master_roster_data['student_id_alt_tc'].apply(
        wf_core_data.utils.extract_alphanumeric
    )
    ### Normalized gender
    logger.info('Constructing normalized gender field')
    gender_map_dict = gender_map['gender_wf'].to_dict()
    master_roster_data['student_gender_wf'] = master_roster_data['student_gender_tc'].apply(
        lambda x: gender_map_dict.get(x, None) if not pd.isna(x) else None
    )
    ### Normalized ethnicity
    logger.info('Constructing normalized ethnicity field')
    student_ethnicity_wf = (
        master_roster_data['student_ethnicity_tc']
        .dropna()
        .explode()
        .to_frame()
        .join(
            ethnicity_map,
            how='left',
            on='student_ethnicity_tc'
        )
        .dropna(subset=['ethnicity_short_name_wf'])
        .groupby(['school_id_tc', 'student_id_tc'])
        .aggregate(student_ethnicity_wf = ('ethnicity_short_name_wf', lambda x: sorted(list(set(x.to_list())))))
    )
    master_roster_data = (
        master_roster_data
        .join(
            student_ethnicity_wf,
            how='left'
        )
    )
    ### Normalized grade
    logger.info('Constructing normalized grade field')
    grade_map_dict = grade_map['grade_short_name_wf'].to_dict()
    def extract_grade_name(row):
        if not pd.isna(row['student_grade_tc']) and slugify.slugify(row['student_grade_tc'], separator='_') in grade_map_dict.keys():
            return grade_map_dict.get(slugify.slugify(row['student_grade_tc'], separator='_'))
        if not pd.isna(row['student_program_tc']) and slugify.slugify(row['student_program_tc'], separator='_') in grade_map_dict.keys():
            return grade_map_dict.get(slugify.slugify(row['student_program_tc'], separator='_'))
        if not pd.isna(row['classroom_name_tc']) and slugify.slugify(row['classroom_name_tc'], separator='_') in grade_map_dict.keys():
            return grade_map_dict.get(slugify.slugify(row['classroom_name_tc'], separator='_'))
        return None
    master_roster_data['student_grade_wf'] = master_roster_data.apply(extract_grade_name, axis=1)
    ## Rearrange rows and columns
    logger.info('Rearranging rows and columns')
    master_roster_data = (
        master_roster_data
        .reindex(columns=[
            'hub_short_name_wf',
            'hub_name_wf',
            'legal_entity_short_name_wf',
            'legal_entity_name_wf',
            'school_short_name_wf',
            'school_name_tc',
            'school_zip_code_tc',
            'school_state',
            'district_id_wida',
            'district_name_wida',
            'school_id_wida',
            'school_name_wida',
            'classroom_id_tc',
            'classroom_short_name_wf',
            'classroom_name_tc',
            'teacher_id_tc',
            'teacher_short_name_wf',
            'teacher_first_name_tc',
            'teacher_last_name_tc',
            'teacher_email_tc',
            'student_first_name_tc',
            'student_last_name_tc',
            'student_middle_name_tc',
            'student_birth_date_tc',
            'student_gender_wf',
            'student_gender_tc',
            'student_ethnicity_wf',
            'student_ethnicity_tc',
            'student_dominant_language_tc',
            'student_household_income_tc',
            'student_grade_wf',
            'student_grade_tc',
            'student_program_tc',
            'student_id_alt_tc',
            'student_id_alt_normalized_tc'
        ])
        .sort_values([
            'hub_short_name_wf',
            'legal_entity_short_name_wf',
            'school_short_name_wf',
            'classroom_short_name_wf',
            'student_last_name_tc',
            'student_first_name_tc',
        ])
    )
    return master_roster_data

def write_master_roster_data_local(
    master_roster_data,
    base_directory,
    subdirectory='master_rosters',
    filename_stem='master_roster',
    filename_suffix=None
):
    if filename_suffix is None:
        filename_suffix = datetime.datetime.now(tz=datetime.timezone.utc).strftime('%Y%m%d')
    logger.info('Filename suffix is {}'.format(filename_suffix))
    output_directory = os.path.join(
        base_directory,
        subdirectory,
        '{}_{}'.format(
            filename_stem,
            filename_suffix
        )
    )
    os.makedirs(output_directory, exist_ok=True)
    logger.info('Writing pickle file')
    master_roster_data.to_pickle(
        os.path.join(
            output_directory,
            '{}_{}.pkl'.format(
                filename_stem,
                filename_suffix
            )
        )
    )
    logger.info('Writing CSV file')
    master_roster_data.to_csv(
        os.path.join(
            output_directory,
            '{}_{}.csv'.format(
                filename_stem,
                filename_suffix
            )
        )
    )



import wf_core_data.rosters.shared_constants
import pandas as pd
import os
import logging

logger = logging.getLogger(__name__)

FOUNTAS_PINNELL_TARGET_COLUMN_NAMES = [
    '## First Name',
    'Last Name',
    'Student ID',
    'DOB',
    'Grade',
    'Gender',
    'Race/Ethnicity',
    'Language',
    'School Lunch Program',
    'Eng Lang Learner',
    'SpEd',
    'Addl Reading Serv',
    'Other Services Programs',
    'Other Services Description',
    'Custom 1',
    'Custom 2'
]

FOUNTAS_PINNELL_GENDER_MAP = {
    'M': 'M',
    'F': 'F',
    'unmatched_value': 'F',
    'na_value': 'F'
}

FOUNTAS_PINNELL_ETHNICITY_MAP = {
    'african_american': 'Black',
    'asian_american': 'Asian/Pacific Islander',
    'hispanic': 'Hispanic',
    'middle_eastern': 'Other',
    'native_american': 'Native American/Alaskan',
    'other': 'Other',
    'pacific_islander': 'Asian/Pacific Islander',
    'white': 'White',
    'unmatched_value': 'Other',
    'na_value': None,
    'multiple_values': 'Multiracial'
}

FOUNTAS_PINNELL_GRADE_NAME_MAP = {
    'EC': 'EC',
    'PK': 'PK',
    'PK_3': 'PK',
    'PK_4': 'PK',
    'K': 'K',
    '1': '1',
    '2': '2',
    '3': '3',
    '4': '4',
    '5': '5',
    '6': '6',
    '7': '7',
    '8': '8',
    '9': '9',
    '10': '10',
    '11': '11',
    '12': '12',
    'unmatched_value': None,
    'na_value': None
}

FOUNTAS_PINNELL_TESTABLE_GRADES = [
    'K',
    '1',
    '2',
    '3',
    '4',
    '5',
    '6',
    '7',
    '8',
    '9',
    '10',
    '11',
    '12'
]

def create_fountas_pinnell_roster_and_write_locally(
    base_directory,
    filename_suffix,
    master_roster_subdirectory='master_rosters',
    master_roster_filename_stem='master_roster',
    fountas_pinnell_roster_subdirectory='fountas_pinnell_rosters',
    fountas_pinnell_roster_filename_stem='fountas_pinnell_roster'
):
    filename = os.path.join(
        base_directory,
        master_roster_subdirectory,
        '{}_{}'.format(
            master_roster_filename_stem,
            filename_suffix
        ),
        '{}_{}.pkl'.format(
            master_roster_filename_stem,
            filename_suffix
        )
    )
    master_roster_data = pd.read_pickle(filename)
    fountas_pinnell_roster_data = wf_core_data.create_fountas_pinnell_roster(
        master_roster_data=master_roster_data
    )
    wf_core_data.write_fountas_pinnell_rosters_local(
        fountas_pinnell_roster_data=fountas_pinnell_roster_data,
        base_directory=base_directory,
        subdirectory=fountas_pinnell_roster_subdirectory,
        filename_stem=fountas_pinnell_roster_filename_stem,
        filename_suffix=filename_suffix
    )

def create_fountas_pinnell_roster(
    master_roster_data
):
    ## Rename fields
    logger.info('Renaming fields')
    fountas_pinnell_roster_data = (
        master_roster_data
        .rename(columns = {
            'student_first_name_tc': '## First Name',
            'student_last_name_tc': 'Last Name'
        })
    )
    ## Create new fields
    ### Student ID
    logger.info('Creating student ID field')
    fountas_pinnell_roster_data['Student ID'] = fountas_pinnell_roster_data.apply(
        lambda row: '{}-{}'.format(
            row.name[0],
            row.name[1]
        ),
        axis=1
    )
    ### Student birth date
    logger.info('Creating birth date field')
    fountas_pinnell_roster_data['DOB'] = fountas_pinnell_roster_data['student_birth_date_tc'].apply(
        lambda x: x.strftime('%m/%d/%Y')
    )
    ### Student gender
    logger.info('Creating gender field')
    fountas_pinnell_roster_data['Gender'] = fountas_pinnell_roster_data['student_gender_wf'].apply(
        lambda x: FOUNTAS_PINNELL_GENDER_MAP.get(x, FOUNTAS_PINNELL_GENDER_MAP.get('unmatched_value')) if pd.notna(x) else FOUNTAS_PINNELL_GENDER_MAP.get('na_value')
    )
    ### Grade
    logger.info('Creating grade field')
    fountas_pinnell_roster_data['Grade'] = fountas_pinnell_roster_data['student_grade_wf'].apply(
        lambda x: FOUNTAS_PINNELL_GRADE_NAME_MAP.get(x, FOUNTAS_PINNELL_GRADE_NAME_MAP.get('unmatched_value')) if pd.notna(x) else FOUNTAS_PINNELL_GRADE_NAME_MAP.get('na_value')
    )
    ### Student ethnicity
    logger.info('Creating ethnicity field')
    def student_race_fountas_pinnell(ethnicity_list):
        if not isinstance(ethnicity_list, list):
            return FOUNTAS_PINNELL_ETHNICITY_MAP.get('na_value')
        if len(ethnicity_list) > 1:
            return FOUNTAS_PINNELL_ETHNICITY_MAP.get('multiple_values')
        return FOUNTAS_PINNELL_ETHNICITY_MAP.get(ethnicity_list[0], FOUNTAS_PINNELL_ETHNICITY_MAP.get('unmatched_value'))
    fountas_pinnell_roster_data['Race/Ethnicity'] = fountas_pinnell_roster_data['student_ethnicity_wf'].apply(student_race_fountas_pinnell)
    ### Language
    fountas_pinnell_roster_data['Language'] = 'English'
    ## Arrange columns and rows
    logger.info('Rearranging columns and rows')
    fountas_pinnell_roster_data = (
        fountas_pinnell_roster_data
        .reindex(columns=(
            wf_core_data.rosters.shared_constants.GROUPING_COLUMN_NAMES +
            FOUNTAS_PINNELL_TARGET_COLUMN_NAMES
        ))
        .sort_values(
            wf_core_data.rosters.shared_constants.GROUPING_COLUMN_NAMES +
            ['Grade', '## First Name', 'Last Name']
        )
    )
    ## Create output
    logger.info('Restriction to testable grades. {} student records before restricting'.format(
        len(fountas_pinnell_roster_data)
    ))
    fountas_pinnell_roster_data = (
        fountas_pinnell_roster_data
        .loc[fountas_pinnell_roster_data['Grade'].isin(FOUNTAS_PINNELL_TESTABLE_GRADES)]
        .copy()
        .reset_index(drop=True)
        .astype('object')
    )
    logger.info('Restricted to testable grades. {} student records after restricting'.format(
        len(fountas_pinnell_roster_data)
    ))
    return fountas_pinnell_roster_data

def write_fountas_pinnell_rosters_local(
    fountas_pinnell_roster_data,
    base_directory,
    subdirectory='fountas_pinnell_rosters',
    filename_stem='fountas_pinnell_roster',
    filename_suffix=None

):
    wf_core_data.rosters.shared_functions.write_rosters_local(
        roster_data=fountas_pinnell_roster_data,
        base_directory=base_directory,
        subdirectory=subdirectory,
        filename_stem=filename_stem,
        filename_suffix=filename_suffix

    )



import wf_core_data.utils
import requests
import pandas as pd
from collections import OrderedDict
import pickle
import json
import datetime
import logging
import os

logger = logging.getLogger(__name__)

class TransparentClassroomClient:
    def __init__(
        self,
        username=None,
        password=None,
        api_token=None,
        url_base='https://www.transparentclassroom.com/api/v1/'
    ):
        self.username = username
        self.password = password
        self.api_token = api_token
        self.url_base = url_base
        if self.api_token is None:
            self.api_token = os.getenv('TRANSPARENT_CLASSROOM_API_TOKEN')
        if self.api_token is None:
            logger.info('Transparent Classroom API token not specified. Attempting to generate token.')
            if self.username is None:
                self.username = os.getenv('TRANSPARENT_CLASSROOM_USERNAME')
            if self.username is None:
                raise ValueError('Transparent Classroom username not specified')
            if self.password is None:
                self.password = os.getenv('TRANSPARENT_CLASSROOM_PASSWORD')
            if self.password is None:
                raise ValueError('Transparent Classroom password not specified')
            json_output = self.transparent_classroom_request(
                'authenticate.json',
                auth=(self.username, self.password)
            )
            self.api_token = json_output['api_token']

    def fetch_and_write_data_local(
        self,
        base_directory,
        pull_datetime=None,
        only_current=False,
        format='dataframe',
        output_directory_stem='transparent_classroom_snapshot',
        all_data_list_filename_stem ='data_tc_list_dict',
        school_data_filename_stem='school_data_tc',
        classroom_data_filename_stem='classroom_data_tc',
        user_data_filename_stem='user_data_tc',
        teacher_default_classroom_data_filename_stem='teacher_default_classroom_data_tc',
        teacher_accessible_classroom_data_filename_stem='teacher_accessible_classroom_data_tc',
        session_data_filename_stem='session_data_tc',
        student_data_filename_stem='student_data_tc',
        student_classroom_data_filename_stem='student_classroom_data_tc',
        student_parent_data_filename_stem='student_parent_data_tc'
    ):
        pull_datetime=wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        data = self.fetch_data(
            pull_datetime=pull_datetime,
            only_current=only_current,
            format=format
        )
        write_data_local(
            data=data,
            base_directory=base_directory,
            only_current=only_current,
            format=format,
            output_directory_stem=output_directory_stem,
            all_data_list_filename_stem=all_data_list_filename_stem,
            school_data_filename_stem=school_data_filename_stem,
            classroom_data_filename_stem=classroom_data_filename_stem,
            user_data_filename_stem=user_data_filename_stem,
            teacher_default_classroom_data_filename_stem=teacher_default_classroom_data_filename_stem,
            teacher_accessible_classroom_data_filename_stem=teacher_accessible_classroom_data_filename_stem,
            session_data_filename_stem=session_data_filename_stem,
            student_data_filename_stem=student_data_filename_stem,
            student_classroom_data_filename_stem=student_classroom_data_filename_stem,
            student_parent_data_filename_stem=student_parent_data_filename_stem
        )

    def fetch_data(
        self,
        pull_datetime=None,
        only_current=False,
        format='dataframe'
    ):
        pull_datetime=wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        logger.info('Fetching all data from Transparent Classroom for all schools and sessions')
        school_data = self.fetch_school_data(
            pull_datetime=pull_datetime,
            format='list'
        )
        school_ids = [school.get('school_id_tc') for school in school_data]
        logger.info('Fetched {} school IDs'.format(len(school_ids)))
        data = {
            'pull_datetime': pull_datetime,
            'schools': school_data,
            'classrooms': list(),
            'users': list(),
            'teachers_default_classrooms': list(),
            'teachers_accessible_classrooms': list(),
            'sessions': list(),
            'students': list(),
            'students_classrooms': list(),
            'students_parents': list()
        }
        for school_id in school_ids:
            data_school= self.fetch_data_school(
                school_id=school_id,
                pull_datetime=pull_datetime,
                only_current=only_current,
                format='list'
            )
            data['classrooms'].extend(data_school['classrooms'])
            data['users'].extend(data_school['users'])
            data['teachers_default_classrooms'].extend(data_school['teachers_default_classrooms'])
            data['teachers_accessible_classrooms'].extend(data_school['teachers_accessible_classrooms'])
            data['sessions'].extend(data_school['sessions'])
            data['students'].extend(data_school['students'])
            data['students_classrooms'].extend(data_school['students_classrooms'])
            data['students_parents'].extend(data_school['students_parents'])
        if format == 'dataframe':
            data['schools'] = convert_school_data_to_df(data['schools'])
            data['classrooms'] = convert_classroom_data_to_df(data['classrooms'])
            data['users'] = convert_user_data_to_df(data['users'])
            data['teachers_default_classrooms'] = convert_teacher_default_classroom_data_to_df(data['teachers_default_classrooms'])
            data['teachers_accessible_classrooms'] = convert_teacher_accessible_classroom_data_to_df(data['teachers_accessible_classrooms'])
            data['sessions'] = convert_session_data_to_df(data['sessions'])
            data['students'] = convert_student_data_to_df(data['students'])
            data['students_classrooms'] = convert_student_classroom_data_to_df(data['students_classrooms'])
            data['students_parents'] = convert_student_parent_data_to_df(data['students_parents'])
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return data

    def fetch_data_school(
        self,
        school_id,
        pull_datetime=None,
        only_current=False,
        format='dataframe'
    ):
        school_id = int(school_id)
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        if only_current:
            logger.info('Fetching only current data from Transparent Classroom for school ID {} for current session'.format(
                school_id
            ))
        else:
            logger.info('Fetching all data from Transparent Classroom for school ID {} for all sessions'.format(
                school_id
            ))
        session_data_school = self.fetch_session_data_school(
            school_id=school_id,
            pull_datetime=pull_datetime,
            format='list'
        )
        classroom_data_school = self.fetch_classroom_data_school(
            school_id=school_id,
            pull_datetime=pull_datetime,
            format='list'
        )
        user_data_school = self.fetch_user_data_school(
            school_id=school_id,
            pull_datetime=pull_datetime,
            format='list'
        )
        student_data_school, student_parent_data_school = self.fetch_student_data_school(
            school_id=school_id,
            pull_datetime=pull_datetime,
            only_current=only_current,
            format='list'
        )
        teacher_default_classroom_data = list()
        teacher_accessible_classroom_data=list()
        for user_datum in user_data_school:
            if 'teacher' in user_datum.get('user_roles_tc', []):
                teacher_default_classroom_data_teacher, teacher_accessible_classroom_data_teacher = self.fetch_teacher_classroom_data_teacher(
                    school_id=school_id,
                    user_id=user_datum['user_id_tc'],
                    pull_datetime=pull_datetime,
                    format='list'
                )
                teacher_default_classroom_data.extend(teacher_default_classroom_data_teacher)
                teacher_accessible_classroom_data.extend(teacher_accessible_classroom_data_teacher)
        if only_current:
            logger.info('Fetching student classroom association data from Transparent Classroom for school ID {} for current session'.format(
                school_id
            ))
        else:
            logger.info('Fetching student classroom association data from Transparent Classroom for school ID {} for each session'.format(
                school_id
            ))
        student_classroom_data = list()
        for session in session_data_school:
            if only_current and not session.get('session_current_tc'):
                continue
            student_classroom_data_session = self.fetch_student_classroom_data_session(
                school_id=school_id,
                session_id=session.get('session_id_tc'),
                pull_datetime=pull_datetime,
                format='list'
            )
            student_classroom_data.extend(student_classroom_data_session)
        data_school = {
            'classrooms': classroom_data_school,
            'users': user_data_school,
            'teachers_default_classrooms': teacher_default_classroom_data,
            'teachers_accessible_classrooms': teacher_accessible_classroom_data,
            'sessions': session_data_school,
            'students': student_data_school,
            'students_classrooms': student_classroom_data,
            'students_parents': student_parent_data_school
        }
        if format == 'dataframe':
            data_school['classrooms'] = convert_classroom_data_to_df(data_school['classrooms'])
            data_school['users'] = convert_user_data_to_df(data_school['users'])
            data_school['teachers_default_classrooms'] = convert_teacher_default_classroom_data_to_df(data_school['teachers_default_classrooms'])
            data_school['teachers_accessible_classrooms'] = convert_teacher_accessible_classroom_data_to_df(data_school['teachers_accessible_classrooms'])
            data_school['sessions'] = convert_session_data_to_df(data_school['sessions'])
            data_school['students'] = convert_student_data_to_df(data_school['students'])
            data_school['students_classrooms'] = convert_student_classroom_data_to_df(data_school['students_classrooms'])
            data_school['students_parents'] = convert_student_parent_data_to_df(data_school['students_parents'])
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return data_school

    def fetch_student_data(
        self,
        school_ids=None,
        pull_datetime=None,
        only_current=False,
        format='dataframe'
    ):
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        if school_ids is None:
            school_ids=self.fetch_school_ids()
        logger.info('Fetching student data from Transparent Classroom for {} schools'.format(len(school_ids)))
        student_data = list()
        student_parent_data = list()
        for school_id in school_ids:
            student_data_school, student_parent_data_school = self.fetch_student_data_school(
                school_id=school_id,
                pull_datetime=pull_datetime,
                only_current=only_current,
                format='list'
            )
            student_data.extend(student_data_school)
            student_parent_data.extend(student_parent_data_school)
        if format == 'dataframe':
            student_data = convert_student_data_to_df(student_data)
            student_parent_data = convert_student_parent_data_to_df(student_parent_data)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return student_data, student_parent_data

    def fetch_student_data_school(
        self,
        school_id,
        pull_datetime=None,
        only_current=False,
        format='dataframe'
    ):
        school_id = int(school_id)
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        logger.info('Fetching student data from Transparent Classroom for school ID {} for all sessions'.format(
            school_id
        ))
        if only_current:
            params={
                'only_current': 'true'
            }
        else:
            params={
                'session_id': 'all'
            }
        json_output = self.transparent_classroom_request(
            'children.json',
            params=params,
            school_id=school_id
        )
        if not isinstance(json_output, list):
            raise ValueError('Received unexpected response from Transparent Classroom: {}'.format(
                json_output
            ))
        student_data_school = list()
        student_parent_data_school = list()
        for datum in json_output:
            student_datum = OrderedDict([
                ('school_id_tc', school_id),
                ('student_id_tc', datum.get('id')),
                ('pull_datetime', pull_datetime),
                ('student_first_name_tc', datum.get('first_name')),
                ('student_middle_name_tc', datum.get('middle_name')),
                ('student_last_name_tc', datum.get('last_name')),
                ('student_birth_date_tc', wf_core_data.utils.to_date(datum.get('birth_date'))),
                ('student_gender_tc', datum.get('gender')),
                ('student_hours_string_tc', datum.get('hours_string')),
                ('student_dominant_language_tc', datum.get('dominant_language')),
                ('student_allergies_tc', datum.get('allergies')),
                ('student_ethnicity_tc', datum.get('ethnicity')),
                ('student_household_income_tc', datum.get('household_income')),
                ('student_approved_adults_string_tc', datum.get('approved_adults_string')),
                ('student_emergency_contacts_string_tc', datum.get('emergency_contacts_string')),
                ('student_program_tc', datum.get('program')),
                ('student_grade_tc', datum.get('grade')),
                ('student_last_day_tc', wf_core_data.utils.to_date(datum.get('last_day'))),
                ('student_exit_reason_tc', datum.get('exit_reason')),
                ('student_id_alt_tc', datum.get('student_id')),
                ('student_notes_tc', datum.get('notes')),
                ('student_exit_survey_id_tc', datum.get('exit_survey_id')),
                ('student_exit_notes_tc', datum.get('exit_notes'))
            ])
            student_data_school.append(student_datum)
            for parent_id in datum.get('parent_ids', []):
                student_parent_datum = OrderedDict([
                    ('school_id_tc', school_id),
                    ('student_id_tc', datum.get('id')),
                    ('pull_datetime', pull_datetime),
                    ('parent_id_tc', parent_id)
                ])
                student_parent_data_school.append(student_parent_datum)
        if format == 'dataframe':
            student_data_school = convert_student_data_to_df(student_data_school)
            student_parent_data_school = convert_student_parent_data_to_df(student_parent_data_school)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return student_data_school, student_parent_data_school

    def fetch_student_classroom_data(
        self,
        school_ids=None,
        session_data=None,
        pull_datetime=None,
        only_current=False,
        format='dataframe'
    ):
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        if school_ids is not None and session_data is not None:
            raise ValueError('Cannot specify both session data and school IDs')
        student_classroom_data = list()
        if session_data is not None:
            if not isinstance(session_data, pd.DataFrame):
                session_data = convert_session_data_to_df(session_data)
            logger.info('Fetching student classroom association data from Transparent Classroom for {} schools'.format(
                len(session_data.index.get_level_values('school_id_tc').unique())
            ))
            for school_id, group_df in session_data.groupby('school_id_tc'):
                # print(group_df)
                session_data_list = group_df.reset_index().to_dict(orient='records')
                # print(session_data_list)
                student_classroom_data_school = self.fetch_student_classroom_data_school(
                    school_id=school_id,
                    session_data_list=session_data_list,
                    pull_datetime=pull_datetime,
                    only_current=only_current,
                    format='list'
                )
                student_classroom_data.extend(student_classroom_data_school)
        else:
            if school_ids is None:
                school_ids=self.fetch_school_ids()
            logger.info('Fetching student classroom association data from Transparent Classroom for {} schools'.format(
                len(school_ids)
            ))
            for school_id in school_ids:
                student_classroom_data_school = self.fetch_student_classroom_data_school(
                    school_id=school_id,
                    session_data_list=None,
                    pull_datetime=pull_datetime,
                    only_current=only_current,
                    format='list'
                )
                student_classroom_data.extend(student_classroom_data_school)
        if format == 'dataframe':
            student_classroom_data = convert_student_classroom_data_to_df(student_classroom_data)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return student_classroom_data

    def fetch_student_classroom_data_school(
        self,
        school_id,
        session_data_list=None,
        pull_datetime=None,
        only_current=False,
        format='dataframe'
    ):
        if school_id is not None:
            school_id = int(school_id)
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        if session_data_list is None:
            session_data_list = self.fetch_session_data_school(
                school_id=school_id,
                pull_datetime=pull_datetime,
                format='list'
            )
        logger.info('Fetching student classroom association data from Transparent Classroom for school ID {}'.format(
            school_id,
        ))
        student_classroom_data_school = list()
        for session in session_data_list:
            if only_current and not session.get('session_current_tc'):
                continue
            student_classroom_data_session = self.fetch_student_classroom_data_session(
                school_id=school_id,
                session_id=session.get('session_id_tc'),
                pull_datetime=pull_datetime,
                format='list'
            )
            student_classroom_data_school.extend(student_classroom_data_session)
        if format == 'dataframe':
            student_classroom_data_school = convert_student_classroom_data_to_df(student_classroom_data_school)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return student_classroom_data_school

    def fetch_student_classroom_data_session(
        self,
        school_id,
        session_id,
        pull_datetime=None,
        format='dataframe'
    ):
        school_id = int(school_id)
        session_id = int(session_id)
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        logger.info('Fetching student classroom association data from Transparent Classroom for school ID {} and session ID {}'.format(
            school_id,
            session_id
        ))
        json_output = self.transparent_classroom_request(
            'children.json',
            params={
                'session_id': session_id
            },
            school_id=school_id
        )
        if not isinstance(json_output, list):
            raise ValueError('Received unexpected response from Transparent Classroom: {}'.format(
                json_output
            ))
        student_classroom_data_session = list()
        for datum in json_output:
            for classroom_id in datum.get('classroom_ids', []):
                student_classroom_datum = OrderedDict([
                    ('school_id_tc', school_id),
                    ('student_id_tc', datum.get('id')),
                    ('pull_datetime', pull_datetime),
                    ('session_id_tc', session_id),
                    ('classroom_id_tc', classroom_id)
                ])
                student_classroom_data_session.append(student_classroom_datum)
        if format == 'dataframe':
            student_classroom_data_session = convert_student_classroom_data_to_df(student_classroom_data_session)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return student_classroom_data_session

    def fetch_session_ids_school(
        self,
        school_id
    ):
        session_data_school = self.fetch_session_data_school(
            school_id=school_id,
            pull_datetime=None,
            format='list'
        )
        session_ids_school = [session.get('session_id_tc') for session in session_data_school]
        return session_ids_school

    def fetch_session_data(
        self,
        school_ids=None,
        pull_datetime=None,
        format='dataframe'
    ):
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        if school_ids is None:
            school_ids=self.fetch_school_ids()
        logger.info('Fetching session data from Transparent Classroom for {} schools'.format(len(school_ids)))
        session_data = list()
        for school_id in school_ids:
            session_data_school = self.fetch_session_data_school(
                school_id=school_id,
                pull_datetime=pull_datetime,
                format='list'
            )
            session_data.extend(session_data_school)
        if format == 'dataframe':
            session_data = convert_session_data_to_df(session_data)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return session_data

    def fetch_session_data_school(
        self,
        school_id,
        pull_datetime=None,
        format='dataframe'
    ):
        school_id = int(school_id)
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        logger.info('Fetching session data from Transparent Classroom for school ID {}'.format(school_id))
        json_output = self.transparent_classroom_request('sessions.json', school_id=school_id)
        if not isinstance(json_output, list):
            raise ValueError('Received unexpected response from Transparent Classroom: {}'.format(
                json_output
            ))
        session_data_school=list()
        for datum in json_output:
            session_datum = OrderedDict([
                ('school_id_tc', school_id),
                ('session_id_tc', datum.get('id')),
                ('pull_datetime', pull_datetime),
                ('session_name_tc', datum.get('name')),
                ('session_start_date_tc', wf_core_data.utils.to_date(datum.get('start_date'))),
                ('session_stop_date_tc', wf_core_data.utils.to_date(datum.get('stop_date'))),
                ('session_current_tc', wf_core_data.utils.to_boolean(datum.get('current'))),
                ('session_inactive_tc', wf_core_data.utils.to_boolean(datum.get('inactive'))),
                ('session_num_children_tc', int(datum.get('children')))
            ])
            session_data_school.append(session_datum)
        if format == 'dataframe':
            session_data_school = convert_session_data_to_df(session_data_school)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return session_data_school

    def fetch_teacher_classroom_data(
        self,
        school_ids=None,
        user_data=None,
        pull_datetime=None,
        format='dataframe'
    ):
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        if school_ids is not None and user_data is not None:
            raise ValueError('Cannot specify both school IDs and user data')
        teacher_default_classroom_data = list()
        teacher_accessible_classroom_data = list()
        if user_data is not None:
            if not isinstance(user_data, pd.DataFrame):
                user_data = convert_user_data_to_df(user_data)
            teacher_data = user_data.loc[user_data['user_roles_tc'].apply(lambda x: 'teacher' in x)]
            logger.info('Fetching teacher classroom association data from Transparent Classroom for {} schools'.format(
                len(teacher_data.index.get_level_values('school_id_tc').unique())
            ))
            for school_id, group_df in teacher_data.groupby('school_id_tc'):
                user_ids = group_df.index.get_level_values('user_id_tc').unique().tolist()
                teacher_default_classroom_data_school, teacher_accessible_classroom_data_school = self.fetch_teacher_classroom_data_school(
                    school_id=school_id,
                    user_ids=user_ids,
                    pull_datetime=pull_datetime,
                    format='list'
                )
                teacher_default_classroom_data.extend(teacher_default_classroom_data_school)
                teacher_accessible_classroom_data.extend(teacher_accessible_classroom_data_school)
        else:
            if school_ids is None:
                school_ids = self.fetch_school_ids()
            logger.info('Fetching teacher classroom association data from Transparent Classroom for {} schools'.format(
                len(school_ids)
            ))
            for school_id in school_ids:
                teacher_default_classroom_data_school, teacher_accessible_classroom_data_school = self.fetch_teacher_classroom_data_school(
                    school_id=school_id,
                    user_ids=None,
                    pull_datetime=pull_datetime,
                    format='list'
                )
                teacher_default_classroom_data.extend(teacher_default_classroom_data_school)
                teacher_accessible_classroom_data.extend(teacher_accessible_classroom_data_school)
        if format == 'dataframe':
            teacher_default_classroom_data = convert_teacher_default_classroom_data_to_df(teacher_default_classroom_data)
            teacher_accessible_classroom_data = convert_teacher_accessible_classroom_data_to_df(teacher_accessible_classroom_data)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return teacher_default_classroom_data, teacher_accessible_classroom_data

    def fetch_teacher_classroom_data_school(
        self,
        school_id,
        user_ids=None,
        pull_datetime=None,
        format='dataframe'
    ):
        school_id = int(school_id)
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        if user_ids is None:
            user_ids = self.fetch_teacher_user_ids_school(school_id=school_id)
        logger.info('Fetching teacher classroom association data from Transparent Classroom for school ID {} for {} teachers'.format(
            school_id,
            len(user_ids)
        ))
        teacher_default_classroom_data_school = list()
        teacher_accessible_classroom_data_school = list()
        for user_id in user_ids:
            teacher_default_classroom_data_teacher, teacher_accessible_classroom_data_teacher = self.fetch_teacher_classroom_data_teacher(
                school_id=school_id,
                user_id=user_id,
                pull_datetime=pull_datetime,
                format='list'
            )
            teacher_default_classroom_data_school.extend(teacher_default_classroom_data_teacher)
            teacher_accessible_classroom_data_school.extend(teacher_accessible_classroom_data_teacher)
        if format == 'dataframe':
            teacher_default_classroom_data_school = convert_teacher_default_classroom_data_to_df(teacher_default_classroom_data_school)
            teacher_accessible_classroom_data_school = convert_teacher_accessible_classroom_data_to_df(teacher_accessible_classroom_data_school)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return teacher_default_classroom_data_school, teacher_accessible_classroom_data_school

    def fetch_teacher_classroom_data_teacher(
        self,
        school_id,
        user_id,
        pull_datetime=None,
        format='dataframe'
    ):
        school_id = int(school_id)
        user_id = int(user_id)
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        logger.info('Fetching teacher classroom association data from Transparent Classroom for school ID {} and user id {}'.format(
            school_id,
            user_id
        ))
        teacher_datum = self.transparent_classroom_request(
            'users/{}.json'.format(user_id),
            school_id=school_id
        )
        if not isinstance(teacher_datum, dict):
            raise ValueError('Received unexpected response from Transparent Classroom: {}'.format(
                teacher_datum
            ))
        teacher_default_classroom_data_teacher = list()
        teacher_accessible_classroom_data_teacher = list()
        if teacher_datum.get('default_classroom_id') is not None:
            teacher_default_classroom_datum = OrderedDict([
                ('school_id_tc', school_id),
                ('user_id_tc', teacher_datum.get('id')),
                ('pull_datetime', pull_datetime),
                ('teacher_default_classroom_id_tc', teacher_datum.get('default_classroom_id'))
            ])
            teacher_default_classroom_data_teacher.append(teacher_default_classroom_datum)
        for accessible_classroom_id in teacher_datum.get('accessible_classroom_ids', []):
            teacher_accessible_classroom_datum = OrderedDict([
                ('school_id_tc', school_id),
                ('user_id_tc', teacher_datum.get('id')),
                ('pull_datetime', pull_datetime),
                ('teacher_accessible_classroom_id_tc', accessible_classroom_id)
            ])
            teacher_accessible_classroom_data_teacher.append(teacher_accessible_classroom_datum)
        if format == 'dataframe':
            teacher_default_classroom_data_teacher = convert_teacher_default_classroom_data_to_df(teacher_default_classroom_data_teacher)
            teacher_accessible_classroom_data_teacher = convert_teacher_accessible_classroom_data_to_df(teacher_accessible_classroom_data_teacher)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return teacher_default_classroom_data_teacher, teacher_accessible_classroom_data_teacher

    def fetch_teacher_user_ids_school(
        self,
        school_id
    ):
        teachers_school = self.fetch_teacher_data_school(
            school_id=school_id,
            pull_datetime=None,
            format='list'
        )
        teacher_user_ids_school = [teacher.get('user_id_tc') for teacher in teachers_school]
        return teacher_user_ids_school

    def fetch_teacher_data(
        self,
        school_ids=None,
        pull_datetime=None,
        format='dataframe'
    ):
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        if school_ids is None:
            school_ids=self.fetch_school_ids()
        logger.info('Fetching teacher data from Transparent Classroom for {} schools'.format(len(school_ids)))
        user_data = self.fetch_user_data(
            school_ids=school_ids,
            pull_datetime=pull_datetime,
            format='list'
        )
        teacher_data = list(filter(lambda user: 'teacher' in user.get('user_roles_tc'), user_data))
        if format == 'dataframe':
            teacher_data = convert_user_data_to_df(teacher_data)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return teacher_data

    def fetch_teacher_data_school(
        self,
        school_id,
        pull_datetime=None,
        format='dataframe'
    ):
        school_id = int(school_id)
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        logger.info('Fetching teacher data from Transparent Classroom for school ID {}'.format(
            school_id
        ))
        user_data_school = self.fetch_user_data_school(
            school_id=school_id,
            pull_datetime=pull_datetime,
            format='list'
        )
        teacher_data_school = list(filter(lambda user: 'teacher' in user.get('user_roles_tc'), user_data_school))
        if format == 'dataframe':
            teacher_data_school = convert_user_data_to_df(teacher_data_school)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return teacher_data_school

    def fetch_parent_data(
        self,
        school_ids=None,
        pull_datetime=None,
        format='dataframe'
    ):
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        if school_ids is None:
            school_ids=self.fetch_school_ids()
        logger.info('Fetching parent data from Transparent Classroom for {} schools'.format(len(school_ids)))
        user_data = self.fetch_user_data(
            school_ids=school_ids,
            pull_datetime=pull_datetime,
            format='list'
        )
        parent_data = list(filter(lambda user: 'parent' in user.get('user_roles_tc'), user_data))
        if format == 'dataframe':
            parent_data = convert_user_data_to_df(parent_data)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return parent_data

    def fetch_parent_data_school(
        self,
        school_id,
        pull_datetime=None,
        format='dataframe'
    ):
        school_id = int(school_id)
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        logger.info('Fetching parent data from Transparent Classroom for school ID {}'.format(
            school_id
        ))
        user_data_school = self.fetch_user_data_school(
            school_id=school_id,
            pull_datetime=pull_datetime,
            format='list'
        )
        parent_data_school = list(filter(lambda user: 'parent' in user.get('user_roles_tc'), user_data_school))
        if format == 'dataframe':
            parent_data_school = convert_user_data_to_df(parent_data_school)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return parent_data_school

    def fetch_user_data(
        self,
        school_ids=None,
        pull_datetime=None,
        format='dataframe'
    ):
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        if school_ids is None:
            school_ids=self.fetch_school_ids()
        logger.info('Fetching user data from Transparent Classroom for {} schools'.format(len(school_ids)))
        user_data = list()
        for school_id in school_ids:
            user_data_school = self.fetch_user_data_school(
                school_id=school_id,
                pull_datetime=pull_datetime,
                format='list'
            )
            user_data.extend(user_data_school)
        if format == 'dataframe':
            user_data = convert_user_data_to_df(user_data)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return user_data

    def fetch_user_data_school(
        self,
        school_id,
        pull_datetime=None,
        format='dataframe'
    ):
        school_id = int(school_id)
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        logger.info('Fetching user data from Transparent Classroom for school ID {}'.format(
            school_id
        ))
        json_output = self.transparent_classroom_request(
            'users.json',
            school_id=school_id
        )
        if not isinstance(json_output, list):
            raise ValueError('Received unexpected response from Transparent Classroom: {}'.format(
                json_output
            ))
        user_data_school = list()
        for datum in json_output:
            user_datum = OrderedDict([
                ('school_id_tc', school_id),
                ('user_id_tc', datum.get('id')),
                ('pull_datetime', pull_datetime),
                ('user_first_name_tc', datum.get('first_name')),
                ('user_last_name_tc', datum.get('last_name')),
                ('user_email_tc', datum.get('email')),
                ('user_roles_tc', datum.get('roles'))
            ])
            user_data_school.append(user_datum)
        if format == 'dataframe':
            user_data_school = convert_user_data_to_df(user_data_school)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return user_data_school

    def fetch_classroom_data(
        self,
        school_ids=None,
        pull_datetime=None,
        format='dataframe'
    ):
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        if school_ids is None:
            school_ids=self.fetch_school_ids()
        logger.info('Fetching classroom data from Transparent Classroom for {} schools'.format(len(school_ids)))
        classroom_data = list()
        for school_id in school_ids:
            classroom_data_school = self.fetch_classroom_data_school(
                school_id=school_id,
                pull_datetime=pull_datetime,
                format='list'
            )
            classroom_data.extend(classroom_data_school)
        if format == 'dataframe':
            classroom_data = convert_classroom_data_to_df(classroom_data)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return classroom_data

    def fetch_classroom_data_school(
        self,
        school_id,
        pull_datetime=None,
        format='dataframe'
    ):
        school_id = int(school_id)
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        logger.info('Fetching classroom data from Transparent Classroom for school ID {}'.format(school_id))
        json_output = self.transparent_classroom_request(
            'classrooms.json',
                params={
                    'show_inactive': True
                },
            school_id=school_id
        )
        if not isinstance(json_output, list):
            raise ValueError('Received unexpected response from Transparent Classroom: {}'.format(
                json_output
            ))
        classroom_data_school=list()
        for datum in json_output:
            classroom_datum = OrderedDict([
                ('school_id_tc', school_id),
                ('classroom_id_tc', datum.get('id')),
                ('pull_datetime', pull_datetime),
                ('classroom_name_tc', datum.get('name')),
                ('classroom_lesson_set_id_tc', datum.get('lesson_set_id')),
                ('classroom_level_tc', datum.get('level')),
                ('classroom_active_tc', wf_core_data.utils.to_boolean(datum.get('active')))
            ])
            classroom_data_school.append(classroom_datum)
        if format == 'dataframe':
            classroom_data_school = convert_classroom_data_to_df(classroom_data_school)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return classroom_data_school

    def fetch_school_ids(self):
        logger.info('Fetching school IDs for all schools')
        school_data = self.fetch_school_data(
            pull_datetime=None,
            format='list'
        )
        school_ids = [school.get('school_id_tc') for school in school_data]
        return school_ids

    def fetch_school_data(
        self,
        pull_datetime=None,
        format='dataframe'
    ):
        pull_datetime = wf_core_data.utils.to_datetime(pull_datetime)
        if pull_datetime is None:
            pull_datetime = datetime.datetime.now(tz=datetime.timezone.utc)
        logger.info('Fetching school data from Transparent Classroom for all schools')
        json_output = self.transparent_classroom_request('schools.json')
        if not isinstance(json_output, list):
            raise ValueError('Received unexpected response from Transparent Classroom: {}'.format(
                json_output
            ))
        school_data=list()
        for datum in json_output:
            if datum.get('type') == 'School':
                school_datum = OrderedDict([
                    ('school_id_tc', datum.get('id')),
                    ('pull_datetime', pull_datetime),
                    ('school_name_tc', datum.get('name')),
                    ('school_address_tc', datum.get('address')),
                    ('school_phone_tc', datum.get('phone')),
                    ('school_time_zone_tc', datum.get('time_zone'))
                ])
                school_data.append(school_datum)
        if format == 'dataframe':
            school_data = convert_school_data_to_df(school_data)
        elif format == 'list':
            pass
        else:
            raise ValueError('Data format \'{}\' not recognized'.format(format))
        return school_data

    def transparent_classroom_request(
        self,
        endpoint,
        params=None,
        school_id=None,
        masquerade_id=None,
        auth=None
    ):
        headers = dict()
        if self.api_token is not None:
            headers['X-TransparentClassroomToken'] = self.api_token
        if school_id is not None:
            headers['X-TransparentClassroomSchoolId'] = str(school_id)
        if masquerade_id is not None:
            headers['X-TransparentClassroomMasqueradeId'] = str(masquerade_id)
        r = requests.get(
            '{}{}'.format(self.url_base, endpoint),
            params=params,
            headers=headers,
            auth=auth
        )
        if r.status_code != 200:
            error_message = 'Transparent Classroom request returned status code {}'.format(r.status_code)
            try:
                if r.json().get('errors') is not None:
                    error_message += '\n{}'.format(json.dumps(r.json().get('errors'), indent=2))
            except:
                pass
            raise Exception(error_message)
        return r.json()

def write_data_local(
    data,
    base_directory,
    only_current=False,
    format='dataframe',
    output_directory_stem='transparent_classroom_snapshot',
    all_data_list_filename_stem ='data_tc_list_dict',
    school_data_filename_stem='school_data_tc',
    classroom_data_filename_stem='classroom_data_tc',
    user_data_filename_stem='user_data_tc',
    teacher_default_classroom_data_filename_stem='teacher_default_classroom_data_tc',
    teacher_accessible_classroom_data_filename_stem='teacher_accessible_classroom_data_tc',
    session_data_filename_stem='session_data_tc',
    student_data_filename_stem='student_data_tc',
    student_classroom_data_filename_stem='student_classroom_data_tc',
    student_parent_data_filename_stem='student_parent_data_tc'
):
    if only_current:
        output_directory_stem = output_directory_stem + '_current'
    # Create local directory
    pull_datetime = data['pull_datetime']
    timestamp_filename_suffix = pull_datetime.strftime('%Y%m%d_%H%M%S')
    output_directory = os.path.join(
        base_directory,
        '{}_{}'.format(
            output_directory_stem,
            timestamp_filename_suffix
        )
    )
    logger.info('Creating directory \'{}\''.format(
        output_directory
    ))
    os.makedirs(output_directory, exist_ok=True)
    if format == 'list':
        with open(
            os.path.join(
                output_directory,
                '{}_{}.pkl'.format(
                    all_data_list_filename_stem,
                    timestamp_filename_suffix
            ))
            ,'wb'
        ) as fp:
            pickle.dump(data, fp)
    elif format == 'dataframe':
        # Write school data in dataframe form to disk
        data['schools'].to_csv(
            os.path.join(
                output_directory,
                '{}_{}.csv'.format(
                    school_data_filename_stem,
                    timestamp_filename_suffix
                )
            ),
            index=False
        )
        data['schools'].to_pickle(os.path.join(
            output_directory,
            '{}_{}.pkl'.format(
                school_data_filename_stem,
                timestamp_filename_suffix
            )
        ))
        # Write classroom data in dataframe form to disk
        data['classrooms'].to_csv(
            os.path.join(
                output_directory,
                '{}_{}.csv'.format(
                    classroom_data_filename_stem,
                    timestamp_filename_suffix
                )
            ),
            index=False
        )
        data['classrooms'].to_pickle(os.path.join(
            output_directory,
            '{}_{}.pkl'.format(
                classroom_data_filename_stem,
                timestamp_filename_suffix
            )
        ))
        # Write user data in dataframe form to disk
        data['users'].to_csv(
            os.path.join(
                output_directory,
                '{}_{}.csv'.format(
                    user_data_filename_stem,
                    timestamp_filename_suffix
                )
            ),
            index=False
        )
        data['users'].to_pickle(os.path.join(
            output_directory,
            '{}_{}.pkl'.format(
                user_data_filename_stem,
                timestamp_filename_suffix
            )
        ))
        # Write teacher default classroom association data in dataframe form to disk
        data['teachers_default_classrooms'].to_csv(
            os.path.join(
                output_directory,
                '{}_{}.csv'.format(
                    teacher_default_classroom_data_filename_stem,
                    timestamp_filename_suffix
                )
            ),
            index=False
        )
        data['teachers_default_classrooms'].to_pickle(os.path.join(
            output_directory,
            '{}_{}.pkl'.format(
                teacher_default_classroom_data_filename_stem,
                timestamp_filename_suffix
            )
        ))
        # Write teacher accessible classroom association data in dataframe form to disk
        data['teachers_accessible_classrooms'].to_csv(
            os.path.join(
                output_directory,
                '{}_{}.csv'.format(
                    teacher_accessible_classroom_data_filename_stem,
                    timestamp_filename_suffix
                )
            ),
            index=False
        )
        data['teachers_accessible_classrooms'].to_pickle(os.path.join(
            output_directory,
            '{}_{}.pkl'.format(
                teacher_accessible_classroom_data_filename_stem,
                timestamp_filename_suffix
            )
        ))
        # Write session data in dataframe form to disk
        data['sessions'].to_csv(
            os.path.join(
                output_directory,
                '{}_{}.csv'.format(
                    session_data_filename_stem,
                    timestamp_filename_suffix
                )
            ),
            index=False
        )
        data['sessions'].to_pickle(os.path.join(
            output_directory,
            '{}_{}.pkl'.format(
                session_data_filename_stem,
                timestamp_filename_suffix
            )
        ))
        # Write student data in dataframe form to disk
        data['students'].to_csv(
            os.path.join(
                output_directory,
                '{}_{}.csv'.format(
                    student_data_filename_stem,
                    timestamp_filename_suffix
                )
            ),
            index=False
        )
        data['students'].to_pickle(os.path.join(
            output_directory,
            '{}_{}.pkl'.format(
                student_data_filename_stem,
                timestamp_filename_suffix
            )
        ))
        # Write student classroom association data in dataframe form to disk
        data['students_classrooms'].to_csv(
            os.path.join(
                output_directory,
                '{}_{}.csv'.format(
                    student_classroom_data_filename_stem,
                    timestamp_filename_suffix
                )
            ),
            index=False
        )
        data['students_classrooms'].to_pickle(os.path.join(
            output_directory,
            '{}_{}.pkl'.format(
                student_classroom_data_filename_stem,
                timestamp_filename_suffix
            )
        ))
        # Write student parent association data in dataframe form to disk
        data['students_parents'].to_csv(
            os.path.join(
                output_directory,
                '{}_{}.csv'.format(
                    student_parent_data_filename_stem,
                    timestamp_filename_suffix
                )
            ),
            index=False
        )
        data['students_parents'].to_pickle(os.path.join(
            output_directory,
            '{}_{}.pkl'.format(
                student_parent_data_filename_stem,
                timestamp_filename_suffix
            )
        ))
    else:
        raise ValueError('Data format \'{}\' not recognized'.format(format))

def convert_school_data_to_df(school_data):
    if len(school_data) == 0:
        return pd.DataFrame()
    school_data_df = pd.DataFrame(
        school_data,
        dtype='object'
    )
    school_data_df['pull_datetime'] = pd.to_datetime(school_data_df['pull_datetime'])
    school_data_df = school_data_df.astype({
        'school_id_tc': 'int',
        'school_name_tc': 'string',
        'school_address_tc': 'string',
        'school_phone_tc': 'string',
        'school_time_zone_tc': 'string'
    })
    school_data_df.set_index('school_id_tc', inplace=True)
    return school_data_df

def convert_classroom_data_to_df(classroom_data):
    if len(classroom_data) == 0:
        return pd.DataFrame()
    classroom_data_df = pd.DataFrame(
        classroom_data,
        dtype='object'
    )
    classroom_data_df['pull_datetime'] = pd.to_datetime(classroom_data_df['pull_datetime'])
    classroom_data_df = classroom_data_df.astype({
        'school_id_tc': 'int',
        'classroom_id_tc': 'int',
        'classroom_name_tc': 'string',
        'classroom_lesson_set_id_tc': 'Int64',
        'classroom_level_tc': 'string',
        'classroom_active_tc': 'bool'
    })
    classroom_data_df.set_index(['school_id_tc', 'classroom_id_tc'], inplace=True)
    return classroom_data_df

def convert_user_data_to_df(user_data):
    if len(user_data) == 0:
        return pd.DataFrame()
    user_data_df = pd.DataFrame(
        user_data,
        dtype='object'
    )
    user_data_df['pull_datetime'] = pd.to_datetime(user_data_df['pull_datetime'])
    user_data_df = user_data_df.astype({
        'school_id_tc': 'int',
        'user_id_tc': 'int',
        'user_first_name_tc': 'string',
        'user_last_name_tc': 'string',
        'user_email_tc': 'string',
    })
    user_data_df.set_index(['school_id_tc', 'user_id_tc'], inplace=True)
    return user_data_df

def convert_teacher_default_classroom_data_to_df(teacher_default_classroom_data):
    if len(teacher_default_classroom_data) == 0:
        return pd.DataFrame()
    teacher_default_classroom_data_df = pd.DataFrame(
        teacher_default_classroom_data,
        dtype='object'
    )
    teacher_default_classroom_data_df['pull_datetime'] = pd.to_datetime(teacher_default_classroom_data_df['pull_datetime'])
    teacher_default_classroom_data_df = teacher_default_classroom_data_df.astype({
        'school_id_tc': 'int',
        'user_id_tc': 'int',
        'teacher_default_classroom_id_tc': 'int'
    })
    teacher_default_classroom_data_df.set_index(['school_id_tc', 'user_id_tc'], inplace=True)
    return teacher_default_classroom_data_df

def convert_teacher_accessible_classroom_data_to_df(teacher_accessible_classroom_data):
    if len(teacher_accessible_classroom_data) == 0:
        return pd.DataFrame()
    teacher_accessible_classroom_data_df = pd.DataFrame(
        teacher_accessible_classroom_data,
        dtype='object'
    )
    teacher_accessible_classroom_data_df['pull_datetime'] = pd.to_datetime(teacher_accessible_classroom_data_df['pull_datetime'])
    teacher_accessible_classroom_data_df = teacher_accessible_classroom_data_df.astype({
        'school_id_tc': 'int',
        'user_id_tc': 'int',
        'teacher_accessible_classroom_id_tc': 'int'
    })
    teacher_accessible_classroom_data_df.set_index(['school_id_tc', 'user_id_tc', 'teacher_accessible_classroom_id_tc'], inplace=True)
    return teacher_accessible_classroom_data_df

def convert_session_data_to_df(session_data):
    if len(session_data) == 0:
        return pd.DataFrame()
    session_data_df = pd.DataFrame(
        session_data,
        dtype='object'
    )
    session_data_df['pull_datetime'] = pd.to_datetime(session_data_df['pull_datetime'])
    session_data_df = session_data_df.astype({
        'school_id_tc': 'int',
        'session_id_tc': 'int',
        'session_name_tc': 'string',
        'session_current_tc': 'bool',
        'session_inactive_tc': 'bool',
        'session_num_children_tc': 'Int64'
    })
    session_data_df.set_index(['school_id_tc', 'session_id_tc'], inplace=True)
    return session_data_df

def convert_student_data_to_df(student_data):
    if len(student_data) == 0:
        return pd.DataFrame()
    student_data_df = pd.DataFrame(
        student_data,
        dtype='object'
    )
    student_data_df['pull_datetime'] = pd.to_datetime(student_data_df['pull_datetime'])
    student_data_df = student_data_df.astype({
            'school_id_tc': 'int',
            'student_id_tc': 'int',
            'student_first_name_tc': 'string',
            'student_middle_name_tc': 'string',
            'student_last_name_tc': 'string',
            'student_birth_date_tc': 'object',
            'student_gender_tc': 'string',
            'student_ethnicity_tc': 'object',
            'student_dominant_language_tc': 'string',
            'student_household_income_tc': 'string',
            'student_grade_tc': 'string',
            'student_program_tc': 'string',
            'student_hours_string_tc': 'string',
            'student_id_alt_tc': 'string',
            'student_allergies_tc': 'string',
            'student_approved_adults_string_tc': 'string',
            'student_emergency_contacts_string_tc': 'string',
            'student_notes_tc': 'string',
            'student_last_day_tc': 'object',
            'student_exit_reason_tc': 'string',
            'student_exit_survey_id_tc': 'Int64',
            'student_exit_notes_tc': 'string'
    })
    student_data_df.set_index(['school_id_tc', 'student_id_tc'], inplace=True)
    return student_data_df

def convert_student_classroom_data_to_df(student_classroom_data):
    if len(student_classroom_data) == 0:
        return pd.DataFrame()
    student_classroom_data_df = pd.DataFrame(
        student_classroom_data,
        dtype='object'
    )
    student_classroom_data_df['pull_datetime'] = pd.to_datetime(student_classroom_data_df['pull_datetime'])
    student_classroom_data_df = student_classroom_data_df.astype({
            'school_id_tc': 'int',
            'student_id_tc': 'int',
            'session_id_tc': 'int',
            'classroom_id_tc': 'int'
    })
    student_classroom_data_df.set_index(['school_id_tc', 'student_id_tc', 'session_id_tc', 'classroom_id_tc'], inplace=True)
    return student_classroom_data_df

def convert_student_parent_data_to_df(student_parent_data):
    if len(student_parent_data) == 0:
        return pd.DataFrame()
    student_parent_data_df = pd.DataFrame(
        student_parent_data,
        dtype='object'
    )
    student_parent_data_df['pull_datetime'] = pd.to_datetime(student_parent_data_df['pull_datetime'])
    student_parent_data_df = student_parent_data_df.astype({
            'school_id_tc': 'int',
            'student_id_tc': 'int',
            'parent_id_tc': 'int'
    })
    student_parent_data_df.set_index(['school_id_tc', 'student_id_tc', 'parent_id_tc'], inplace=True)
    return student_parent_data_df
